"""Module for customized observables"""
from __future__ import annotations

from itertools import combinations
from itertools import combinations_with_replacement

import numpy as np
from numpy import empty
from sklearn.utils.validation import check_is_fitted

from ..common import validate_input
from ._base import BaseObservables


class CustomObservables(BaseObservables):
    """
    A class to map state variables using custom observables.

    This class allows the user to specify a list of functions that map state variables
    to observables. The identity map is automatically included. It can be configured to
    include or exclude self-interaction terms.

    Attributes:
        observables (list of callable): List of functions mapping state variables to
            observables. Univariate functions are applied to each state variable,
            and multivariable functions are applied to combinations of state
            variables. The identity map is automatically included in this list.
        observable_names (list of callable, optional): List of functions mapping from
            names of state variables to names of observables. For example,
            the observable name lambda x: f"{x}^2" would correspond to the function
            x^2. If None, the names "f0(...)", "f1(...)", ... will be used. Default
            is None.
        interaction_only (bool, optional): If True, omits self-interaction terms.
            Function evaluations of the form f(x,x) and f(x,y,x) will be omitted,
            but those of the form f(x,y) and f(x,y,z) will be included. If False,
            all combinations will be included. Default is True.
        n_input_features_ (int): Number of input features.
        n_output_features_ (int): Number of output features.
    """

    def __init__(self, observables, observable_names=None, interaction_only=True):
        """
        Initialize a CustomObservables instance.

        Args:
            observables (list of callable): List of functions mapping state variables
                to observables. Univariate functions are applied to each state
                variable, and multivariable functions are applied to combinations of
                state variables. The identity map is automatically included in this
                list.
            observable_names (list of callable, optional): List of functions mapping
                from names of state variables to names of observables. For example,
                the observable name lambda x: f"{x}^2" would correspond to the
                function x^2. If None, the names "f0(...)", "f1(...)", ... will
                be used. Default is None.
            interaction_only (bool, optional): If True, omits self-interaction terms.
                Function evaluations of the form f(x,x) and f(x,y,x) will be omitted,
                but those of the form f(x,y) and f(x,y,z) will be included. If False,
                all combinations will be included. Default is True.
        """
        super(CustomObservables, self).__init__()
        self.observables = [identity, *observables]
        if observable_names and (len(observables) != len(observable_names)):
            raise ValueError(
                "observables and observable_names must have the same length"
            )
        self.observable_names = observable_names
        self.interaction_only = interaction_only
        self.include_state = True

    def fit(self, x, y=None):
        """
        Fit the model to the measurement data.

        This method calculates the number of input and output features and generates
        default values for 'observable_names' if necessary. It also prepares the
        measurement matrix for data transformation.

        Args:
            x (array-like, shape (n_samples, n_input_features)): Measurement data to be
                fitted.
            y (None): This is a dummy parameter added for compatibility with sklearn's
                API. Default is None.

        Returns:
            self (CustomObservables): This method returns the fitted instance.
        """
        x = validate_input(x)
        n_samples, n_features = x.shape

        n_output_features = 0
        for f in self.observables:
            n_args = f.__code__.co_argcount
            n_output_features += len(
                list(self._combinations(n_features, n_args, self.interaction_only))
            )

        self.n_input_features_ = n_features
        self.n_output_features_ = n_output_features
        self.n_consumed_samples = 0

        if self.observable_names is None:
            self.observable_names = list(
                map(
                    lambda i: (lambda *x: "f" + str(i) + "(" + ",".join(x) + ")"),
                    range(len(self.observables)),
                )
            )

        # First map is the identity
        self.observable_names.insert(0, identity_name)

        # since the first map is identity
        self.measurement_matrix_ = np.zeros(
            (self.n_input_features_, self.n_output_features_)
        )
        self.measurement_matrix_[
            : self.n_input_features_, : self.n_input_features_
        ] = np.eye(self.n_input_features_)

        return self

    def transform(self, x):
        """
        Apply custom transformations to data, computing observables.

        This method applies the user-defined observables functions to the input data,
        effectively transforming the state variables into observable ones.

        Args:
            x (array-like, shape (n_samples, n_input_features)): The measurement data
                to be transformed.

        Returns:
            x_transformed (array-like, shape (n_samples, n_output_features)): The
                transformed data, i.e., the computed observables.
        """
        check_is_fitted(self, "n_input_features_")
        check_is_fitted(self, "n_output_features_")
        x = validate_input(x)

        n_samples, n_features = x.shape

        if n_features != self.n_input_features_:
            raise ValueError("x.shape[1] does not match n_input_features_")

        x_transformed = empty((n_samples, self.n_output_features_), dtype=x.dtype)
        observables_idx = 0
        for f in self.observables:
            for c in self._combinations(
                self.n_input_features_, f.__code__.co_argcount, self.interaction_only
            ):
                x_transformed[:, observables_idx] = f(*[x[:, j] for j in c])
                observables_idx += 1

        return x_transformed

    def get_feature_names(self, input_features=None):
        """
        Get the names of the output features.

        This method returns the names of the output features as defined by the
        observable functions. If names for the input features are provided, they are
        used in the output feature names. Otherwise, default names ("x0", "x1", ...,
        "xn_input_features") are used.

        Args:
            input_features (list of string, length n_input_features, optional):
                String names for input features, if available. By default, the names
                "x0", "x1", ... ,"xn_input_features" are used.

        Returns:
            output_feature_names (list of string, length n_output_features):
                Output feature names.
        """
        check_is_fitted(self, "n_input_features_")
        if input_features is None:
            input_features = [f"x{i}" for i in range(self.n_input_features_)]
        else:
            if len(input_features) != self.n_input_features_:
                raise ValueError(
                    "input_features must have n_input_features_ "
                    f"({self.n_input_features_}) elements"
                )

        feature_names = []
        for i, f in enumerate(self.observables):
            feature_names.extend(
                [
                    self.observable_names[i](*[input_features[j] for j in c])
                    for c in self._combinations(
                        self.n_input_features_,
                        f.__code__.co_argcount,
                        self.interaction_only,
                    )
                ]
            )

        return feature_names

    @staticmethod
    def _combinations(n_features, n_args, interaction_only):
        """
        Get the combinations of features to be passed to observable functions.

        This static method generates all possible combinations or combinations with
        replacement (depending on the `interaction_only` flag) of features that are to
        be passed to the observable functions. The combinations are represented as
        tuples of indices.

        Args:
            n_features (int): The total number of features.
            n_args (int): The number of arguments that the observable function accepts.
            interaction_only (bool): If True, combinations of the same feature
                (self-interactions) are omitted. If False, all combinations including
                self-interactions are included.

        Returns:
            iterable of tuples: An iterable over all combinations of feature indices
            to be passed to the observable functions.
        """
        comb = combinations if interaction_only else combinations_with_replacement
        return comb(range(n_features), n_args)


def identity(x):
    """Identity map."""
    return x


def identity_name(x):
    """Name for identity map."""
    return str(x)



"""module for nonlinear schrodinger equation"""
from __future__ import annotations

import numpy as np
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from pykoopman.common.examples import rk4
from scipy.fft import fft
from scipy.fft import fftfreq
from scipy.fft import ifft


class nlse:
    """
    nonlinear schrodinger equation

    iu_t + 0.5u_xx + u*|u|^2 = 0

    periodic B.C. PDE is solved with Spectral methods using FFT
    """

    def __init__(self, n, dt, L=2 * np.pi):
        self.n_states = n
        # assert self.u0.size == self.n_states, 'check the size of initial
        # condition and mesh size n'

        dk = 2 * np.pi / L
        self.k = fftfreq(self.n_states, 1.0 / self.n_states) * dk
        self.dt = dt

    def sys(self, t, x, u):
        """the RHS for the governing equation using FFT"""
        xk = fft(x)

        # 4/3 truncation rule
        # dealiasing due to triple nonlinearity
        # note: you could do zero-padding to improve memory
        # efficiency
        xk[self.n_states // 4 : 3 * self.n_states // 4] = 0j
        x = ifft(xk)

        yk = (-self.k**2 * xk.ravel() / 2) * 1j
        y = ifft(yk) + 1j * abs(x) ** 2 * x + u
        return y

    def simulate(self, x0, n_int, n_sample):
        # n_traj = x0.shape[1]
        x = x0
        u = np.zeros((n_int, 1), dtype=complex)
        X = np.zeros((n_int // n_sample, self.n_states), dtype=complex)
        t = 0
        j = 0
        t_list = []
        for step in range(n_int):
            t += self.dt
            y = rk4(0, x, u[step], self.dt, self.sys)
            if (step + 1) % n_sample == 0:
                X[j] = y
                j += 1
                t_list.append(t)
            x = y
        return X, np.array(t_list)

    def collect_data_continuous(self, x0):
        """
        collect training data pairs - continuous sense.

        given x0, with shape (n_dim, n_traj), the function
        returns dx/dt with shape (n_dim, n_traj)
        """

        n_traj = x0.shape[0]
        u = np.zeros((n_traj, 1))
        X = x0
        Y = []
        for i in range(n_traj):
            y = self.sys(0, x0[i], u[i])
            Y.append(y)
        Y = np.vstack(Y)
        return X, Y

    def collect_one_step_data_discrete(self, x0):
        """
        collect training data pairs - discrete sense.

        given x0, with shape (n_dim, n_traj), the function
        returns system state x1 after self.dt with shape
        (n_dim, n_traj)
        """

        n_traj = x0.shape[0]
        X = x0
        Y = []
        for i in range(n_traj):
            y, _ = self.simulate(x0[i], n_int=1, n_sample=1)
            # for j in range(int(delta_t // self.dt)):
            #     y = rk4(0, x, u[:, i], self.dt, self.sys)
            #     x = y
            Y.append(y)
        Y = np.vstack(Y)
        return X, Y

    def collect_one_trajectory_data(self, x0, n_int, n_sample):
        x = x0
        y, _ = self.simulate(x, n_int, n_sample)
        return y

    def visualize_data(self, x, t, X):
        plt.figure(figsize=(6, 6))
        ax = plt.axes(projection=Axes3D.name)
        for i in range(X.shape[0]):
            ax.plot(x, abs(X[i]), zs=t[i], zdir="t", label="time = " + str(i * self.dt))
        # plt.legend(loc='best')
        ax.view_init(elev=35.0, azim=-65, vertical_axis="y")
        ax.set(ylabel=r"$mag. of u(x,t)$", xlabel=r"$x$", zlabel=r"time $t$")
        plt.title("Nonlinear schrodinger equation (Kutz et al., Complexity, 2018)")
        plt.show()

    def visualize_state_space(self, X):
        u, s, vt = np.linalg.svd(X, full_matrices=False)
        # this is a pde problem so the number of snapshots are smaller than dof
        pca_1_r, pca_1_i = np.real(u[:, 0]), np.imag(u[:, 0])
        pca_2_r, pca_2_i = np.real(u[:, 1]), np.imag(u[:, 1])
        pca_3_r, pca_3_i = np.real(u[:, 2]), np.imag(u[:, 2])

        plt.figure(figsize=(6, 6))
        plt.semilogy(s)
        plt.xlabel("number of SVD terms")
        plt.ylabel("singular values")
        plt.title("PCA singular value decays")
        plt.show()

        plt.figure(figsize=(6, 6))
        ax = plt.axes(projection=Axes3D.name)
        ax.plot3D(pca_1_r, pca_2_r, pca_3_r, "k-o")
        ax.set(xlabel="pc1", ylabel="pc2", zlabel="pc3")
        plt.title("PCA visualization (real)")
        plt.show()

        plt.figure(figsize=(6, 6))
        ax = plt.axes(projection=Axes3D.name)
        ax.plot3D(pca_1_i, pca_2_i, pca_3_i, "k-o")
        ax.set(xlabel="pc1", ylabel="pc2", zlabel="pc3")
        plt.title("PCA visualization (imag)")
        plt.show()


if __name__ == "__main__":
    n = 512
    x = np.linspace(-15, 15, n, endpoint=False)
    u0 = 2.0 / np.cosh(x)
    # u0 = u0.reshape(-1,1)
    n_int = 10000
    n_snapshot = 80  # in the original paper, it is 20, but I think too small
    dt = np.pi / n_int
    n_sample = n_int // n_snapshot

    model = nlse(n, dt=dt, L=30)
    X, t = model.simulate(u0, n_int, n_sample)

    # usage: visualize the data in physical space
    model.visualize_data(x, t, X)

    # usage: visualize the data in state space
    model.visualize_state_space(X)

    print(X.shape)
    print(t[1] - t[0])

    # usage: collect continuous data pair: x and dx/dt
    x0_array = np.vstack([u0, u0, u0])
    X, Y = model.collect_data_continuous(x0_array)

    print(X.shape)
    print(Y.shape)

    # usage: collect discrete data pair
    x0_array = np.vstack([u0, u0, u0])
    X, Y = model.collect_one_step_data_discrete(x0_array)

    print(X.shape)
    print(Y.shape)

    # usage: collect one trajectory data
    X = model.collect_one_trajectory_data(u0, n_int, n_sample)
    print(X.shape)



"""module for dmd"""
# from warnings import warn
from __future__ import annotations

import numpy as np
from pydmd import DMDBase
from pydmd.dmdbase import DMDTimeDict
from pydmd.utils import compute_svd
from pydmd.utils import compute_tlsq
from scipy.linalg import eig
from scipy.linalg import sqrtm
from sklearn.utils.validation import check_is_fitted

from ._base import BaseRegressor


class PyDMDRegressor(BaseRegressor):
    """
    PyDMDRegressor is a wrapper for `pydmd` regressors.

    This class provides a wrapper for the `pydmd` regressor. The details about
    `pydmd` can be found in the reference:

    Demo, N., Tezzele, M., & Rozza, G. (2018). PyDMD: Python dynamic mode decomposition.
    Journal of Open Source Software, 3(22), 530.
    <https://joss.theoj.org/papers/10.21105/joss.00530.pdf>`_

    Args:
        regressor (DMDBase): A regressor instance from DMDBase in `pydmd`.
        tikhonov_regularization (bool or None, optional): Indicates if Tikhonov
        regularization should be applied. Defaults to None.

    Attributes:
        tlsq_rank (int): Rank for truncation in TLSQ method. If 0, no noise reduction
            is computed. If positive, it is used for SVD truncation.
        svd_rank (int): Rank for truncation. If 0, optimal rank is computed and used
            for truncation. If positive integer, it is used for truncation. If float
            between 0 and 1, the rank is the number of the biggest singular values
            that are needed to reach the 'energy' specified by `svd_rank`. If -1, no
            truncation is computed.
        forward_backward (bool): If True, the low-rank operator is computed like in
            fbDMD.
        tikhonov_regularization (bool or None, optional): If None, no regularization
            is applied. If float, it is used as the Tikhonov regularization parameter.
        flag_xy (bool): If True, the regressor is operating on multiple trajectories
            instead of just one.
        n_samples_ (int): Number of samples.
        n_input_features_ (int): Number of features, i.e., the dimension of phi.
        _unnormalized_modes (ndarray): Raw DMD V with each column as one DMD mode.
        _state_matrix_ (ndarray): DMD state transition matrix.
        _reduced_state_matrix_ (ndarray): Reduced DMD state transition matrix.
        _eigenvalues_ (ndarray): Identified Koopman lambda.
        _eigenvectors_ (ndarray): Identified Koopman eigenvectors.
        _coef_ (ndarray): Weight vectors of the regression problem. Corresponds to
            either [A] or [A,B].
        C (ndarray): Matrix that maps psi to the input features.
    """

    def __init__(self, regressor, tikhonov_regularization=None):
        """
        Initializes a PyDMDRegressor instance.

        Args:
            regressor (DMDBase): A regressor instance from DMDBase in `pydmd`.
            tikhonov_regularization (bool or None, optional): Indicates if Tikhonov
                regularization should be applied. Defaults to None.

        Raises:
            ValueError: If regressor is not a subclass of DMDBase from pydmd.
        """
        if not isinstance(regressor, DMDBase):
            raise ValueError("regressor must be a subclass of DMDBase from pydmd.")
        self.regressor = regressor
        # super(PyDMDRegressor, self).__init__(regressor)
        self.tlsq_rank = regressor._tlsq_rank
        self.svd_rank = regressor._Atilde._svd_rank
        self.forward_backward = regressor._Atilde._forward_backward
        self.tikhonov_regularization = tikhonov_regularization
        self.flag_xy = False
        self._ur = None

    def fit(self, x, y=None, dt=1):
        """
        Fit the PyDMDRegressor model according to the given training data.

        Args:
            x (np.ndarray): Measurement data input. Should be of shape (n_samples,
                n_features).
            y (np.ndarray, optional): Measurement data output to be fitted. Should be
                of shape (n_samples, n_features). Defaults to None.
            dt (float, optional): Time interval between `x` and `y`. Defaults to 1.

        Returns:
            self : Returns the instance itself.
        """

        self.n_samples_, self.n_input_features_ = x.shape

        if y is None:
            # single trajectory
            self.flag_xy = False
            X = x[:-1].T
            Y = x[1:].T
        else:
            # multiple segments of trajectories
            self.flag_xy = True
            X = x.T
            Y = y.T

        X, Y = compute_tlsq(X, Y, self.tlsq_rank)
        U, s, V = compute_svd(X, self.svd_rank)

        if self.tikhonov_regularization is not None:
            _norm_X = np.linalg.norm(X)
        else:
            _norm_X = 0

        atilde = self._least_square_operator(
            U, s, V, Y, self.tikhonov_regularization, _norm_X
        )
        if self.forward_backward:
            # b stands for "backward"
            bU, bs, bV = compute_svd(Y, svd_rank=len(s))
            atilde_back = self._least_square_operator(
                bU, bs, bV, X, self.tikhonov_regularization, _norm_X
            )
            atilde = sqrtm(atilde @ np.linalg.inv(atilde_back))

        # - V, lamda, eigenvectors
        self._coef_ = atilde
        self._state_matrix_ = atilde
        [self._eigenvalues_, self._eigenvectors_] = eig(self._state_matrix_)

        # self._coef_ = U @ atilde @ U.conj().T
        # self._state_matrix_ = self._coef_
        # self._reduced_state_matrix_ = atilde
        # [self._eigenvalues_, self._eigenvectors_] = eig(self._reduced_state_matrix_)
        self._ur = U
        self._unnormalized_modes = self._ur @ self._eigenvectors_

        self._tmp_compute_psi = np.linalg.pinv(self.unnormalized_modes)

        # self.C = np.linalg.inv(self._eigenvectors_) @ U.conj().T
        # self._modes_ = U.dot(self._eigenvectors_)

        return self

    def predict(self, x):
        """
        Predict the future values based on the input measurement data.

        Args:
            x (np.ndarray): Measurement data upon which to base the prediction.
                Should be of shape (n_samples, n_features).

        Returns:
            np.ndarray: Predicted values of `x` one timestep in the future. The shape
                is (n_samples, n_features).
        """
        if x.ndim == 1:
            x = x.reshape(1, -1)
        check_is_fitted(self, "coef_")
        y = np.linalg.multi_dot([self.ur, self._coef_, self.ur.conj().T, x.T]).T
        return y

    def _compute_phi(self, x_col):
        """
        Compute the `phi(x)` value given `x`.

        Args:
            x_col (np.ndarray): Input data, if one-dimensional it will be reshaped
                to (-1, 1).

        Returns:
            np.ndarray: Computed `phi(x)` value.
        """
        if x_col.ndim == 1:
            x_col = x_col.reshape(-1, 1)
        phi = self.ur.T @ x_col
        return phi

    def _compute_psi(self, x_col):
        """
        Compute the `psi(x)` value given `x`.

        Args:
            x_col (np.ndarray): Input data, if one-dimensional it will be reshaped
                to (-1, 1).

        Returns:
            np.ndarray: Value of Koopman eigenfunction psi at x.
        """

        # compute psi - one column if x is a row
        if x_col.ndim == 1:
            x_col = x_col.reshape(-1, 1)
        psi = self._tmp_compute_psi @ x_col
        return psi

    def _set_initial_time_dictionary(self, time_dict):
        """
        Sets the initial values for `time_dict` and `original_time`.
        Typically called in `fit()` and not used again afterwards.

        Args:
            time_dict (dict): Initial time dictionary for this DMD instance. Must
                contain the keys "t0", "tend", and "dt".

        Raises:
            ValueError: If the time_dict does not contain the keys "t0", "tend" and
                "dt" or if it contains more than these keys.
        """

        if not ("t0" in time_dict and "tend" in time_dict and "dt" in time_dict):
            raise ValueError(
                'time_dict must contain the keys "t0", ' '"tend" and "dt".'
            )
        if len(time_dict) > 3:
            raise ValueError(
                'time_dict must contain only the keys "t0", ' '"tend" and "dt".'
            )

        self._original_time = DMDTimeDict(dict(time_dict))
        self._dmd_time = DMDTimeDict(dict(time_dict))

    def _least_square_operator(self, U, s, V, Y, tikhonov_regularization, _norm_X):
        """
        Calculates the least square estimation 'A' using the provided parameters.

        Args:
            U (numpy.ndarray): Left singular vectors, shape (n_features, svd_rank).
            s (numpy.ndarray): Singular values, shape (svd_rank, ).
            V (numpy.ndarray): Right singular vectors, shape (n_features, svd_rank).
            Y (numpy.ndarray): Measurement data for prediction, shape (n_samples,
                n_features).
            tikhonov_regularization (bool or NoneType): Tikhonov parameter for
                regularization. If `None`, no regularization is applied, if `float`,
                it is used as the :math:`\\lambda` tikhonov parameter.
            _norm_X (numpy.ndarray): Norm of `X` for Tikhonov regularization, shape
                (n_samples, n_features).

        Returns:
            numpy.ndarray: The least square estimation 'A', shape (svd_rank, svd_rank).
        """

        if tikhonov_regularization is not None:
            s = (s**2 + tikhonov_regularization * _norm_X) * np.reciprocal(s)
        A = np.linalg.multi_dot([U.T.conj(), Y, V]) * np.reciprocal(s)
        return A

    @property
    def coef_(self):
        """
        The weight vectors of the regression problem.

        This method checks if the regressor is fitted before returning the coefficient.

        Returns:
            numpy.ndarray: The coefficient matrix.

        Raises:
            NotFittedError: If the regressor is not fitted yet.
        """
        check_is_fitted(self, "_coef_")
        return self._coef_

    @property
    def state_matrix_(self):
        """
        The DMD state transition matrix.

        This method checks if the regressor is fitted before returning the state matrix.

        Returns:
            numpy.ndarray: The state transition matrix.

        Raises:
            NotFittedError: If the regressor is not fitted yet.
        """
        check_is_fitted(self, "_state_matrix_")
        return self._state_matrix_

    @property
    def eigenvalues_(self):
        """
        The identified Koopman eigenvalues.

        This method checks if the regressor is fitted before returning the eigenvalues.

        Returns:
            numpy.ndarray: The Koopman eigenvalues.

        Raises:
            NotFittedError: If the regressor is not fitted yet.
        """
        check_is_fitted(self, "_eigenvalues_")
        return self._eigenvalues_

    @property
    def eigenvectors_(self):
        """
        The identified Koopman eigenvectors.

        This method checks if the regressor is fitted before returning the eigenvectors.

        Returns:
            numpy.ndarray: The Koopman eigenvectors.

        Raises:
            NotFittedError: If the regressor is not fitted yet.
        """
        check_is_fitted(self, "_eigenvectors_")
        return self._eigenvectors_

    @property
    def unnormalized_modes(self):
        """
        The raw DMD V with each column as one DMD mode.

        This method checks if the regressor is fitted before returning the unnormalized
            modes.

        Returns:
            numpy.ndarray: The unnormalized modes.

        Raises:
            NotFittedError: If the regressor is not fitted yet.
        """
        check_is_fitted(self, "_unnormalized_modes")
        return self._unnormalized_modes

    @property
    def ur(self):
        """
        The left singular vectors 'U'.

        This method checks if the regressor is fitted before returning 'U'.

        Returns:
            numpy.ndarray: The left singular vectors 'U'.

        Raises:
            NotFittedError: If the regressor is not fitted yet.
        """
        check_is_fitted(self, "_ur")
        return self._ur



from __future__ import annotations

from .cqgle import cqgle
from .examples import advance_linear_system
from .examples import drss
from .examples import Linear2Ddynamics
from .examples import lorenz
from .examples import rev_dvdp
from .examples import rk4
from .examples import slow_manifold
from .examples import torus_dynamics
from .examples import vdp_osc
from .ks import ks
from .nlse import nlse
from .validation import check_array
from .validation import drop_nan_rows
from .validation import validate_input
from .vbe import vbe

__all__ = [
    "check_array",
    "drop_nan_rows",
    "validate_input",
    "drss",
    "advance_linear_system",
    "torus_dynamics",
    "lorenz",
    "vdp_osc",
    "rk4",
    "rev_dvdp",
    "Linear2Ddynamics",
    "slow_manifold",
    "nlse",
    "vbe",
    "cqgle",
    "ks",
]



"""module for extended dmd with control"""
from __future__ import annotations

import numpy as np
from sklearn.utils.validation import check_is_fitted

from ._base import BaseRegressor

# TODO: add support for time delay observables, so we will
#       have n_consumption_.


class EDMDc(BaseRegressor):
    """Module for Extended DMD with control (EDMDc) regressor.

    Aims to determine the system matrices A, B, C that satisfy y' = Ay + Bu and x = Cy,
    where y' is the time-shifted observable with y0 = phi(x0) and u is the control
    input. B and C are the unknown control and measurement matrices, respectively.

    The objective functions, \\|Y'-AY-BU\\|_F and \\|X-CY\\|_F, are minimized using
    least-squares regression and singular value decomposition.

    See the following reference for more details:
        Korda, M. and Mezic, I. "Linear predictors for nonlinear dynamical systems:
        Koopman operator meets model predictive control." Automatica, Vol. 93, 149–160.
        <https://www.sciencedirect.com/science/article/abs/pii/S000510981830133X>

    Attributes:
        coef_ (numpy.ndarray):
            Weight vectors of the regression problem. Corresponds to either [A] or
            [A,B].
        state_matrix_ (numpy.ndarray):
            Identified state transition matrix A of the underlying system.
        control_matrix_ (numpy.ndarray):
            Identified control matrix B of the underlying system.
        projection_matrix_ (numpy.ndarray):
            Projection matrix into low-dimensional subspace of shape (n_input_features
            +n_control_features, svd_rank).
        projection_matrix_output_ (numpy.ndarray):
            Projection matrix into low-dimensional subspace of shape (n_input_features
            +n_control_features, svd_output_rank).
    """

    def __init__(self):
        """Initialize the EDMDc regressor."""
        pass

    def fit(self, x, y=None, u=None, dt=None):
        """Fit the EDMDc regressor to the given data.

        Args:
            x (numpy.ndarray):
                Measurement data to be fit.
            y (numpy.ndarray, optional):
                Time-shifted measurement data to be fit. Defaults to None.
            u (numpy.ndarray, optional):
                Time series of external actuation/control. Defaults to None.
            dt (scalar, optional):
                Discrete time-step. Defaults to None.

        Returns:
            self: Fitted EDMDc instance.
        """
        self.n_samples_, self.n_input_features_ = x.shape
        if y is None:
            X1 = x[:-1, :]
            X2 = x[1:, :]
        else:
            X1 = x
            X2 = y

        if u.ndim == 1:
            if len(u) > X1.shape[0]:
                u = u[:-1]
            C = u[np.newaxis, :]
        else:
            if u.shape[0] > X1.shape[0]:
                u = u[:-1, :]
            C = u
        self.n_control_features_ = C.shape[1]

        self._fit_with_unknown_b(X1, X2, C)
        return self

    def _fit_with_unknown_b(self, X1, X2, U):
        """Fit the EDMDc regressor with unknown control matrix B.

        Args:
            X1 (numpy.ndarray):
                Measurement data given as input.
            X2 (numpy.ndarray):
                Measurement data given as target.
            U (numpy.ndarray):
                Time series of external actuation/control.
        """
        Nlift = X1.shape[1]
        W = X2.T
        V = np.vstack([X1.T, U.T])
        VVt = V @ V.T
        WVt = W @ V.T
        M = WVt @ np.linalg.pinv(VVt)  # Matrix [A B]
        self._state_matrix_ = M[0:Nlift, 0:Nlift]
        self._control_matrix_ = M[0:Nlift, Nlift:]
        self._coef_ = M

        # Compute Koopman V, eigenvectors, lamda
        [self._eigenvalues_, self._eigenvectors_] = np.linalg.eig(self.state_matrix_)
        self._unnormalized_modes = self._eigenvectors_
        self._ur = np.eye(self.n_input_features_)
        self._tmp_compute_psi = np.linalg.inv(self._eigenvectors_)

    def predict(self, x, u):
        """Predict the next timestep based on the given data.

        Args:
            x (numpy.ndarray):
                Measurement data upon which to base prediction.
            u (numpy.ndarray):
                Time series of external actuation/control.

        Returns:
            y (numpy.ndarray):
                Prediction of x one timestep in the future.
        """
        check_is_fitted(self, "coef_")
        y = x @ self.state_matrix_.T + u @ self.control_matrix_.T
        return y

    def _compute_phi(self, x_col):
        """Compute psi(x) given x.

        Args:
            x_col (numpy.ndarray):
                Input data x.

        Returns:
            psi (numpy.ndarray):
                Value of psi(x).
        """
        if x_col.ndim == 1:
            x_col = x_col.reshape(-1, 1)
        phi = self._ur.T @ x_col
        return phi

    def _compute_psi(self, x_col):
        """Compute psi(x) given x.

        Args:
            x_col (numpy.ndarray):
                Input data x.

        Returns:
            psi (numpy.ndarray):
                Value of psi(x).
        """
        # compute psi - one column if x is a row
        if x_col.ndim == 1:
            x_col = x_col.reshape(-1, 1)
        psi = self._tmp_compute_psi @ x_col
        return psi

    @property
    def coef_(self):
        """Weight vectors of the regression problem. Corresponds to either [A] or
        [A,B]."""
        check_is_fitted(self, "_coef_")
        return self._coef_

    @property
    def state_matrix_(self):
        """Identified state transition matrix A of the underlying system.

        Returns:
            state_matrix (numpy.ndarray):
                State transition matrix A.
        """
        check_is_fitted(self, "_state_matrix_")
        return self._state_matrix_

    @property
    def control_matrix_(self):
        """Identified control matrix B of the underlying system.

        Returns:
            control_matrix (numpy.ndarray):
                Control matrix B.
        """
        check_is_fitted(self, "_control_matrix_")
        return self._control_matrix_

    @property
    def eigenvalues_(self):
        """Identified Koopman lambda.

        Returns:
            eigenvalues (numpy.ndarray):
                Koopman eigenvalues.
        """
        check_is_fitted(self, "_eigenvalues_")
        return self._eigenvalues_

    @property
    def eigenvectors_(self):
        """Identified Koopman eigenvectors.

        Returns:
            eigenvectors (numpy.ndarray):
                Koopman eigenvectors.
        """
        check_is_fitted(self, "_eigenvectors_")
        return self._eigenvectors_

    @property
    def unnormalized_modes(self):
        """Identified Koopman eigenvectors.

        Returns:
            unnormalized_modes (numpy.ndarray):
                Koopman eigenvectors.
        """
        check_is_fitted(self, "_unnormalized_modes")
        return self._unnormalized_modes

    @property
    def ur(self):
        """Matrix U that is part of the SVD.

        Returns:
            ur (numpy.ndarray):
                Matrix U.
        """
        check_is_fitted(self, "_ur")
        return self._ur



"""module for handling a ensemble of x-x' pair.

Manual changes are made to add support to complex numeric data
"""
from __future__ import annotations

from sklearn.base import BaseEstimator
from sklearn.base import clone
from sklearn.base import TransformerMixin
from sklearn.compose import TransformedTargetRegressor


class EnsembleBaseRegressor(TransformedTargetRegressor):
    """
    This class serves as a wrapper for PyKoopman regressors that utilize ensemble or
    non-consecutive training data.

    `EnsembleBaseRegressor` inherits from `TransformedTargetRegressor` and checks
    whether the provided regressor object implements the `fit` and `predict` methods.

    Attributes:
        regressor (sklearn.base.BaseEstimator): A regressor object that implements
            `fit` and `predict` methods.
        func (function): A function to apply to the target `y` before passing it to
            the `fit` method. The function must return a 2-dimensional array.
            If `func` is `None`, the identity function is used.
        inverse_func (function): A function to apply to the prediction of the
            regressor. This function is used to return predictions to the same space
            as the original training labels. It must return a 2-dimensional array.

    Raises:
        AttributeError: If the regressor does not have a callable `fit` or
            `predict` method.
        ValueError: If both `transformer` and functions `func`/`inverse_func`
            are set, or if 'func' is provided while 'inverse_func' is not.

    Note:
        This class does not implement the `fit` method on its own, instead, it checks
        the methods of the provided regressor object and raises an AttributeError if
        the required methods are not present or not callable. It also performs some
        pre-processing on the target values `y` before fitting the regressor, and
        provides additional checks and warnings for the transformer and inverse
        functions.
    """

    def __init__(self, regressor, func, inverse_func):
        super().__init__(regressor=regressor, func=func, inverse_func=inverse_func)
        if not hasattr(regressor, "fit") or not callable(getattr(regressor, "fit")):
            raise AttributeError("regressor does not have a callable fit method")
        if not hasattr(regressor, "predict") or not callable(
            getattr(regressor, "predict")
        ):
            raise AttributeError("regressor does not have a callable predict method")

    def fit(self, X, y, **fit_params):
        """
        Fits the model according to the given training data.

        Args:
            X (array-like or sparse matrix of shape (n_samples, n_features)):
                Training vector, where `n_samples` is the number of samples and
                `n_features` is the number of features.
            y (array-like of shape (n_samples,)): Target values.
            **fit_params (dict): Additional parameters passed to the `fit` method of
                the underlying regressor.

        Returns:
            self: The fitted estimator.

        Raises:
            ValueError: If 'transformer' and functions 'func'/'inverse_func' are both
                set, or if 'func' is provided while 'inverse_func' is not.

        Note:
            This method transforms the target `y` before fitting the regressor and
            performs additional checks and warnings for the transformer and inverse
            functions.
        """

        # transformers are designed to modify X which is 2d dimensional, we
        # need to modify y accordingly.

        self._training_dim = y.ndim
        if y.ndim == 1:
            y_2d = y.reshape(-1, 1)
        else:
            y_2d = y
        self._fit_transformer(y_2d)

        # transform y and convert back to 1d array if needed
        y_trans = self.transformer_.transform(y_2d)
        # FIXME: a FunctionTransformer can return a 1D array even when validate
        # is set to True. Therefore, we need to check the number of dimension
        # first.
        if y_trans.ndim == 2 and y_trans.shape[1] == 1:
            y_trans = y_trans.squeeze(axis=1)

        if self.regressor is None:
            from sklearn.linear_model import LinearRegression

            self.regressor_ = LinearRegression()
        else:
            self.regressor_ = clone(self.regressor)

        self.regressor_.fit(X, y_trans, **fit_params)

        if hasattr(self.regressor_, "feature_names_in_"):
            self.feature_names_in_ = self.regressor_.feature_names_in_

        return self

    def _fit_transformer(self, y):
        """
        Checks the transformer and fits it.

        This method creates the default transformer if necessary, fits it, and
        performs additional inverse checks on a subset (optional).

        Args:
            y (array-like): The target values.

        Raises:
            ValueError: If both 'transformer' and functions 'func'/'inverse_func'
                are set, or if 'func' is provided while 'inverse_func' is not.

        Note:
            The method does not currently pass 'sample_weight' to the transformer.
            However, if the transformer starts using 'sample_weight', the code should
            be modified accordingly. During the consideration of the 'sample_prop'
            feature, this is also a good use case to consider.
        """
        if self.transformer is not None and (
            self.func is not None or self.inverse_func is not None
        ):
            raise ValueError(
                "'transformer' and functions 'func'/'inverse_func' cannot both be set."
            )
        elif self.transformer is not None:
            self.transformer_ = clone(self.transformer)
        else:
            if self.func is not None and self.inverse_func is None:
                raise ValueError(
                    "When 'func' is provided, 'inverse_func' must also be provided"
                )
            self.transformer_ = FunctionTransformer(
                func=self.func,
                inverse_func=self.inverse_func,
                validate=True,
                check_inverse=self.check_inverse,
            )
        # XXX: sample_weight is not currently passed to the
        # transformer. However, if transformer starts using sample_weight, the
        # code should be modified accordingly. At the time to consider the
        # sample_prop feature, it is also a good use case to be considered.
        self.transformer_.fit(y)
        # if self.check_inverse:
        #     idx_selected = slice(None, None, max(1, y.shape[0] // 10))
        #     y_sel = _safe_indexing(y, idx_selected)
        #     y_sel_t = self.transformer_.transform(y_sel)
        #     if not np.allclose(y_sel, self.transformer_.inverse_transform(y_sel_t)):
        #         warnings.warn(
        #             "The provided functions or transformer are"
        #             " not strictly inverse of each other. If"
        #             " you are sure you want to proceed regardless"
        #             ", set 'check_inverse=False'",
        #             UserWarning,
        #         )


class FunctionTransformer(TransformerMixin, BaseEstimator):
    """Constructs a transformer from an arbitrary callable.

    This class forwards its X (and optionally y) arguments to a user-defined function
    or function object and returns the result of this function. This is useful for
    stateless transformations such as taking the log of frequencies, doing custom
    scaling, etc.

    Note: If a lambda is used as the function, then the resulting transformer will
        not be pickleable.

    Attributes:
        func (callable): The callable to use for the transformation. This will be
            passed the same arguments as transform, with args and kwargs forwarded.
            If func is None, then func will be the identity function.
        inverse_func (callable): The callable to use for the inverse transformation.
            This will be passed the same arguments as inverse transform, with args
            and kwargs forwarded. If inverse_func is None, then inverse_func will be
            the identity function.
        validate (bool): Indicate that the input X array should be checked before
            calling func. The default is False.
        accept_sparse (bool): Indicate that func accepts a sparse matrix as input.
            The default is False.
        check_inverse (bool): Whether to check that or func followed by inverse_func
            leads to the original inputs. The default is True.
        kw_args (dict): Dictionary of additional keyword arguments to pass to func.
        inv_kw_args (dict): Dictionary of additional keyword arguments to pass to
            inverse_func.
        n_input_features_ (int): Number of features seen during fit. Defined only
            when validate=True.
        feature_names_in_ (ndarray): Names of features seen during fit. Defined only
            when validate=True and X has feature names that are all strings.

    Examples:
        >>> import numpy as np
        >>> from sklearn.preprocessing import FunctionTransformer
        >>> transformer = FunctionTransformer(np.log1p)
        >>> X = np.array([[0, 1], [2, 3]])
        >>> transformer.transform(X)
        array([[0.       , 0.6931...],
               [1.0986..., 1.3862...]])
    """

    def __init__(
        self,
        func=None,
        inverse_func=None,
        *,
        validate=False,
        accept_sparse=False,
        check_inverse=True,
        kw_args=None,
        inv_kw_args=None,
    ):
        """Initialize the FunctionTransformer instance.

        Args:
            func (callable, optional): The callable to use for the transformation.
                This will be passed the same arguments as transform, with args and
                kwargs forwarded. If func is None, then
                func will be the identity function. Defaults to None.
            inverse_func (callable, optional): The callable to use for the inverse
                transformation. This will be passed the same arguments as inverse
                transform, with args and kwargs forwarded. If inverse_func is None, then
                inverse_func will be the identity function. Defaults to None.
            validate (bool, optional): Indicate that the input X array should be
                checked before calling func. Defaults to False.
            accept_sparse (bool, optional): Indicate that func accepts a sparse matrix
                as input. Defaults to False.
            check_inverse (bool, optional): Whether to check that func followed by
                inverse_func leads to the original inputs. Defaults to True.
            kw_args (dict, optional): Dictionary of additional keyword arguments to
                pass to func. Defaults to None.
            inv_kw_args (dict, optional): Dictionary of additional keyword arguments
                to pass to inverse_func. Defaults to None.
        """
        self.func = func
        self.inverse_func = inverse_func
        self.validate = validate
        self.accept_sparse = accept_sparse
        self.check_inverse = check_inverse
        self.kw_args = kw_args
        self.inv_kw_args = inv_kw_args

    def _check_input(self, X, *, reset):
        """Checks the input X. If validation is enabled, it validates the data.

        Args:
            X (array-like): Input data to be checked/validated.
            reset (bool): Flag indicating whether to reset the validation.

        Returns:
            array-like: The original input data, possibly validated if `validate`
                attribute is set to True.
        """
        # if self.validate:
        #     return self._validate_data(X, accept_sparse=self.accept_sparse,
        #     reset=reset)
        return X

    def _check_inverse_transform(self, X):
        """Checks if the provided functions are the inverse of each other.

        Selects a subset of X and performs a round trip transformation: forward
        transform followed by inverse transform. Raises a warning if the round trip
        does not return the original inputs.

        Args:
            X (array-like): Input data to be checked for inverse transform consistency.
        """
        idx_selected = slice(None, None, max(1, X.shape[0] // 100))
        # X_round_trip = self.inverse_transform(self.transform(X[idx_selected]))
        self.inverse_transform(self.transform(X[idx_selected]))
        # if not _allclose_dense_sparse(X[idx_selected], X_round_trip):
        #     warnings.warn(
        #         "The provided functions are not strictly"
        #         " inverse of each other. If you are sure you"
        #         " want to proceed regardless, set"
        #         " 'check_inverse=False'.",
        #         UserWarning,
        #     )

    def fit(self, X, y=None):
        """Fits transformer by checking X.

        If ``validate`` is ``True``, ``X`` will be checked. Also checks if the provided
        functions are the inverse of each other if `check_inverse` is set to True.

        Args:
            X (array-like): The data to fit. Shape should be (n_samples, n_features).
            y (None, optional): Ignored. Not used, present here for API consistency by
                convention.

        Returns:
            FunctionTransformer: The fitted transformer.
        """
        X = self._check_input(X, reset=True)
        if self.check_inverse and not (self.func is None or self.inverse_func is None):
            self._check_inverse_transform(X)
        return self

    def transform(self, X):
        """Transforms X using the forward function.

        Args:
            X (array-like): The data to transform. Shape should be (n_samples,
                n_features).

        Returns:
            array-like: Transformed data with same shape as input.
        """
        X = self._check_input(X, reset=False)
        return self._transform(X, func=self.func, kw_args=self.kw_args)

    def inverse_transform(self, X):
        """Transforms X using the inverse function.

        Args:
            X (array-like): The data to inverse transform. Shape should be
                (n_samples, n_features).

        Returns:
            array-like: Inverse transformed data with the same shape as input.
        """
        # if self.validate:
        #     X = check_array(X, accept_sparse=self.accept_sparse)
        return self._transform(X, func=self.inverse_func, kw_args=self.inv_kw_args)

    def _transform(self, X, func=None, kw_args=None):
        """Applies the given function to the data X.

        Args:
            X (array-like): The data to transform. Shape should be (n_samples,
                n_features).
            func (callable, optional): The function to apply. If None, identity
                function is used.
            kw_args (dict, optional): Additional arguments to pass to the function.

        Returns:
            array-like: Transformed data with the same shape as input.
        """
        if func is None:
            func = _identity

        return func(X, **(kw_args if kw_args else {}))

    def __sklearn_is_fitted__(self):
        """Return True since FunctionTransfomer is stateless."""
        return True

    def _more_tags(self):
        return {"no_validation": not self.validate, "stateless": True}


def _identity(X):
    """The identity function."""
    return X



"""Tests for pykoopman.koopman_continuous methods."""
from __future__ import annotations

import pytest
from numpy.testing import assert_allclose
from pydmd import DMD
from pykoopman import KoopmanContinuous
from pykoopman import observables
from pykoopman import regression
from pykoopman.differentiation import Derivative
from sklearn.utils.validation import check_is_fitted


@pytest.mark.parametrize(
    "data",
    [pytest.lazy_fixture("data_random"), pytest.lazy_fixture("data_random_complex")],
)
def test_derivative_integration(data):
    x = data

    diff = Derivative(kind="finite_difference", k=1)
    dmd = DMD(svd_rank=2)
    model = KoopmanContinuous(differentiator=diff, regressor=dmd)

    model.fit(x)
    check_is_fitted(model)


def test_havok_prediction(data_lorenz):
    t, x, dt = data_lorenz

    n_delays = 99
    TDC = observables.TimeDelay(delay=1, n_delays=n_delays)
    HAVOK = regression.HAVOK(svd_rank=15)
    Diff = Derivative(kind="finite_difference", k=2)
    model = KoopmanContinuous(observables=TDC, differentiator=Diff, regressor=HAVOK)
    model.fit(x[:, 0], dt=dt)

    known_external_input = model.regressor.forcing_signal

    # one step prediction
    xpred_one_step = model.predict(x[: n_delays + 1, 0], dt, u=known_external_input[0])
    assert_allclose(x[n_delays + 1, 0], xpred_one_step, atol=1e-3)

    # simulate: mult steps prediction
    xpred = model.simulate(
        x=x[: n_delays + 1, 0], t=t[n_delays:] - t[n_delays], u=known_external_input
    )

    assert_allclose(xpred[50], 3.54512034, atol=1e-3)



"""module for random fourier features observables"""
from __future__ import annotations

import numpy as np
from sklearn.utils.validation import check_is_fitted

from ..common import validate_input
from ._base import BaseObservables


class RandomFourierFeatures(BaseObservables):
    """
    Random Fourier Features for observables.

    This class applies the random Fourier features method for kernel approximation.
    It can include the system state in the kernel function. It uses the
    Gaussian kernel by default.

    Args:
        include_state (bool, optional): If True, includes the system state. Defaults to
            True.
        gamma (float, optional): The scale of the Gaussian kernel. Defaults to 1.0.
        D (int, optional): The number of random samples in Monte Carlo approximation.
            Defaults to 100.
        random_state (int, None, optional): The seed of the random number for repeatable
            experiments. Defaults to None.

    Attributes:
        include_state (bool): If True, includes the system state.
        gamma (float): The scale of the Gaussian kernel.
        D (int): The number of random samples in Monte Carlo approximation.
        random_state (int, None): The seed of the random number for repeatable
            experiments.
        measurement_matrix_ (numpy.ndarray): A row feature vector right multiply with
            `measurement_matrix_` will return the system state.
        n_input_features_ (int): Dimension of input features, e.g., system state.
        n_output_features_ (int): Dimension of transformed/output features, e.g.,
            observables.
        w (numpy.ndarray): The frequencies randomly sampled for random fourier features.
    """

    def __init__(self, include_state=True, gamma=1.0, D=100, random_state=None):
        """
        Initialize the RandomFourierFeatures class with given parameters.

        Args:
            include_state (bool, optional): If True, includes the system state.
                Defaults to True.
            gamma (float, optional): The scale of the Gaussian kernel. Defaults to 1.0.
            D (int, optional): The number of random samples in Monte Carlo
                approximation. Defaults to 100.
            random_state (int or None, optional): The seed of the random number
                for repeatable experiments. Defaults to None.
        """
        super(RandomFourierFeatures, self).__init__()
        self.include_state = include_state
        self.gamma = gamma
        self.D = D
        self.random_state = random_state

    def fit(self, x, y=None):
        """
        Set up observable.

        Args:
            x (numpy.ndarray): Measurement data to be fit. Shape (n_samples,
                n_input_features_).
            y (numpy.ndarray, optional): Time-shifted measurement data to be fit.
                Defaults to None.

        Returns:
            self: Returns a fitted RandomFourierFeatures instance.
        """
        x = validate_input(x)
        np.random.seed(self.random_state)
        self.n_consumed_samples = 0

        self.n_input_features_ = x.shape[1]
        # although we have double the output dim, the convergence
        # rate is described in only self.n_components
        self.n_output_features_ = 2 * self.D

        if self.include_state is True:
            self.n_output_features_ += self.n_input_features_

        # 1. generate (n_feature, n_component) random w
        self.w = np.sqrt(2.0 * self.gamma) * np.random.normal(
            0, 1, [self.n_input_features_, self.D]
        )

        # 3. get the C to map back to state
        if self.include_state:
            self.measurement_matrix_ = np.zeros(
                (self.n_input_features_, self.n_output_features_)
            )
            self.measurement_matrix_[
                : self.n_input_features_, : self.n_input_features_
            ] = np.eye(self.n_input_features_)
        else:
            # we have to transform the data x in order to find a matrix by fitting
            # z = np.zeros((x.shape[0], self.n_output_features_))
            # z[:,:x.shape[1]] = x
            # z[:,x.shape[1]:] = self._rff_lifting(x)
            z = self._rff_lifting(x)
            self.measurement_matrix_ = np.linalg.lstsq(z, x)[0].T

        return self

    def transform(self, x):
        """
        Evaluate observable at `x`.

        Args:
            x (numpy.ndarray): Measurement data to be fit. Shape (n_samples,
                n_input_features_).

        Returns:
            y (numpy.ndarray): Evaluation of observables at `x`. Shape (n_samples,
                n_output_features_).
        """

        check_is_fitted(self, "n_input_features_")
        z = np.zeros((x.shape[0], self.n_output_features_))
        z_rff = self._rff_lifting(x)
        if self.include_state:
            z[:, : x.shape[1]] = x
            z[:, x.shape[1] :] = z_rff
        else:
            z = z_rff

        return z

    def get_feature_names(self, input_features=None):
        """
        Return names of observables.

        Args:
            input_features (list of string of length n_features, optional):
                Default list is "x0", "x1", ..., "xn", where n = n_features.

        Returns:
            output_feature_names (list of string of length n_output_features):
                Returns a list of observable names.
        """

        check_is_fitted(self, "n_input_features_")

        if input_features is None:
            input_features = [f"x{i}" for i in range(self.n_input_features_)]
        else:
            if len(input_features) != self.n_input_features_:
                raise ValueError(
                    "input_features must have n_input_features_ "
                    f"({self.n_input_features_}) elements"
                )

        if self.include_state:
            # very easy to make mistake... python pass list by reference OMG
            output_features = input_features[:]
        else:
            output_features = []
        output_features += [f"cos(w_{i}'x)/sqrt({self.D})" for i in range(self.D)] + [
            f"sin(w_{i}'x)/sqrt({self.D})" for i in range(self.D)
        ]

        return output_features

    def _rff_lifting(self, x):
        """
        Core algorithm that computes random Fourier features.

        This method uses the `cos` and `sin` transformations to get random Fourier
            features.

        Args:
            x (numpy.ndarray): System state.

        Returns:
            z_rff (numpy.ndarray): Random Fourier features evaluated on `x`. Shape
                (n_samples, n_output_features_).
        """

        # 2. get the feature vector z
        xw = np.dot(x, self.w)
        z_rff = np.hstack([np.cos(xw), np.sin(xw)])
        z_rff *= 1.0 / np.sqrt(self.D)
        return z_rff



"""Module for implementing Pan-Duraisamy modes selection algorithm"""
from __future__ import annotations

import numpy as np
from matplotlib import pyplot as plt
from prettytable import PrettyTable
from pykoopman.koopman import Koopman
from sklearn.linear_model import enet_path

from ._base_analyzer import BaseAnalyzer
from ._pruned_koopman import PrunedKoopman


class ModesSelectionPAD21(BaseAnalyzer):
    """Koopman V selection algorithm from Pan, et al.,JFM (2021).

    Aims to extract a low-dimensional Koopman invariant subspace
    in a model-agnostic way, i.e., applies to any algorithms for
    approximating the Koopman operator.

    See the following reference for more details:
        Pan, S., Arnold-Medabalimi, N., & Duraisamy, K. (2021).
        Sparsity-promoting algorithms for the discovery of informative
        Koopman-invariant subspaces. Journal of Fluid Mechanics, 917.
        <https://www.cambridge.org/core/journals/journal-of-fluid-mechanics/
        article/sparsitypromoting-algorithms-for-the-discovery-of-informative-
        koopmaninvariant-subspaces/F52F03ED181023369A661EF09B57281A>

    Parameters:
        model (Koopman): An instance of `pykoopman.koopman.Koopman`.
        validate_data_traj (list): A list of dictionaries that contains validation
            trajectories. Each dictionary has two keys: `t` and `x`, which correspond to
             time stamps and data matrix.
        truncation_threshold (float): When sweeping the `alpha` in the sparse linear
            regression solver over the data, we consider any term with an absolute
            coefficient smaller than the truncation threshold as zero. Default is 1e-3.
        max_terms_allowed (int): The maximum number of terms used to perform sparse
            linear regression. It can be inferred by the Q-R plot. Default is 10.
        plot (bool): True if the figures should be plotted. Default is False.

    Attributes:
        L (int): Total number of terms considered in sparse linear regression.
        dir (str): The path where figures are saved.
        eigenfunction_on_traj_total_top_k (numpy.ndarray): Evaluations of the best k
            eigenfunctions evaluated on all of the validation trajectories.
        small_to_large_error_eigen_index (numpy.ndarray): Indices of eigenmodes from
            best to worst in terms of linear evolution error.
        sweep_index_list (list): A list of a list of bool. It is the final result of
            sweeping in the sparse linear regression. It contains which V are
            selected at a certain `alpha`.
        validate_data_traj (list): A list of dictionaries that contains validation
            trajectories. Each dictionary has two keys: `t` and `x`, which correspond
            to time stamps and data matrix.
        truncation_threshold (float): When sweeping the `alpha` in the sparse linear
            regression solver over the data, we consider any term with an absolute
            coefficient smaller than the truncation threshold as zero.
    """

    def __init__(
        self,
        model: Koopman,
        validate_data_traj: list,
        truncation_threshold=1e-3,
        max_terms_allowed=10,
        plot=False,
    ):
        """Initialize the ModesSelectionPAD21 object.

        Args:
            model (Koopman): An instance of `pykoopman.koopman.Koopman`.
            validate_data_traj (list): A list of dictionaries that contains
                validation trajectories. Each dictionary should have the keys 't'
                and 'x', corresponding to time stamps and data matrix, respectively.
            truncation_threshold (float, optional): When sweeping the alpha in the
                sparse linear regression solver over the data, we consider any term
                with an absolute coefficient smaller than the truncation threshold as
                zero. Defaults to 1e-3.
            max_terms_allowed (int, optional): The maximum number of terms used to
                perform sparse linear regression. It can be inferred by the Q-R plot.
                Defaults to 10.
            plot (bool, optional): Set to True to plot the figures. Defaults to False.
        """

        super().__init__(model)

        self.validate_data_traj = validate_data_traj
        self.truncation_threshold = truncation_threshold
        self.dir = "/"

        if not isinstance(validate_data_traj, list):
            raise NotImplementedError("validate_data_traj should be a list.")

        # loop over each validation trajectory
        Q_i = []
        for validate_data_one_traj in validate_data_traj:
            validate_data = validate_data_one_traj["x"]
            validate_time = validate_data_one_traj["t"]

            # 1. residual of linearity equation
            linear_residual_list = self._compute_phi_minus_phi_evolved(
                validate_time, validate_data
            )

            # 1.1 normalization factor for each psi
            eigenfunction_evaluated_on_traj = self.eigenfunction(validate_data.T).T

            tmp = np.abs(eigenfunction_evaluated_on_traj) ** 2  # pointwise square only
            dt_arr = np.diff(validate_time, prepend=validate_time[0] - validate_time[1])
            tmp = np.dot(dt_arr, tmp) / (validate_time[-1] - validate_time[0])
            normal_constant = np.sqrt(tmp)

            # 1.2 normalized residual and pick the maximized one over t
            tmp = [
                np.max(np.abs(tmp) / normal_constant[i])
                for i, tmp in enumerate(linear_residual_list)
            ]
            Q_i.append(tmp)

        # compute the mean Q_i for all of the trajectories
        Q_i_mean = np.array(Q_i).mean(axis=0)

        # sort Q to get i_1 to i_L
        self.small_to_large_error_eigen_index = np.argsort(Q_i_mean)[
            : max_terms_allowed + 1
        ]

        # loop over each validation trajectory - for the second time
        R_i = []
        for validate_data_one_traj in validate_data_traj:
            validate_data = validate_data_one_traj["x"]
            eigenfunction_evaluated_on_traj = self.eigenfunction(validate_data.T).T

            # get reconstruction error with increasing number of V
            R_i_each = []
            for k in range(1, max_terms_allowed + 1):
                eigenfunction_evaluated_on_traj_top_k = eigenfunction_evaluated_on_traj[
                    :, self.small_to_large_error_eigen_index[:k]
                ]
                sparse_measurement_matrix = np.linalg.lstsq(
                    eigenfunction_evaluated_on_traj_top_k, validate_data
                )[0]
                residual = (
                    eigenfunction_evaluated_on_traj_top_k @ sparse_measurement_matrix
                    - validate_data
                )
                normalized_err_top_k = np.linalg.norm(residual) / np.linalg.norm(
                    validate_data
                )
                R_i_each.append(normalized_err_top_k)
            R_i_each = np.array(R_i_each)
            R_i.append(R_i_each)
        R_i_mean = np.array(R_i).mean(axis=0)

        # print out the Q-R
        QR_table = PrettyTable()

        QR_table.field_names = ["Index", "Eigenvalue", "Q", "R"]
        tmp = self.eigenvalues_discrete[self.small_to_large_error_eigen_index]
        for i in range(len(R_i_mean)):
            QR_table.add_row(
                [
                    self.small_to_large_error_eigen_index[i],
                    tmp[i],
                    Q_i_mean[i],
                    R_i_mean[i],
                ]
            )
        print(QR_table)

        # prepare top max k selected eigentraj
        eigenfunction_evaluated_on_traj_total = np.vstack(
            [self.eigenfunction(tmp1["x"].T).T for tmp1 in validate_data_traj]
        )
        self.eigenfunction_on_traj_total_top_k = eigenfunction_evaluated_on_traj_total[
            :, self.small_to_large_error_eigen_index[: max_terms_allowed + 1]
        ]

        if plot:
            fig = plt.figure(figsize=(6, 6))
            ax1 = fig.add_subplot(111)
            ax2 = ax1.twinx()
            ax1.plot(
                range(1, len(Q_i_mean) + 1),
                Q_i_mean[self.small_to_large_error_eigen_index],
                "b-^",
                label="Max relative psi error",
            )
            ax1.set_xlabel("Number of eigenmodes included", fontsize=16)
            ax1.set_yscale("log")
            ax1.set_ylabel(
                "Max linear evolving normalized error",
                color="b",
                fontsize=16,
            )
            ax2.plot(
                np.arange(1, len(R_i_mean) + 1),
                R_i_mean,
                "r-o",
                label="Reconstruction normalized error",
            )
            ax2.set_ylabel("Reconstruction normalized error", color="r", fontsize=16)
            ax2.set_yscale("log")
            plt.grid("both")
            # annotate the lamda
            for i in range(len(Q_i_mean)):
                ax1.text(
                    i,
                    Q_i_mean[self.small_to_large_error_eigen_index][i],
                    "{:.2f}".format(
                        self.eigenvalues_discrete[
                            self.small_to_large_error_eigen_index
                        ][i]
                    ),
                    size=10,
                    rotation=25,
                )
            plt.tight_layout()
            plt.show()

    def sweep_among_best_L_modes(
        self,
        L: int,
        ALPHA_RANGE=np.logspace(-3, 1, 100),
        MAX_ITER=1e5,
        save_figure=True,
        plot=True,
    ):
        """Perform sweeping among the best L modes using multi-task elastic net.

        Computes multi-task elastic net over a list of alpha values and saves the
        coefficients for each path.

        Parameters:
            L (int): The number of eigenmodes considered for sparse linear regression.
            ALPHA_RANGE (numpy.ndarray, optional): An array of alpha values on which to
                perform sparse linear regression. Defaults to np.logspace(-3, 1, 100).
            MAX_ITER (int, optional): Maximum iterations allowed in the coordinate
                descent algorithm. Defaults to 1e5.
            save_figure (bool, optional): True if the figures should be saved. Defaults
                to True.
            plot (bool, optional): True if the figures should be plotted. Defaults to
                True.
        """

        self.L = L
        # options
        TOL = 1e-12
        L1_RATIO = 0.99

        phi_tilde = self.eigenfunction_on_traj_total_top_k[:, :L]
        X = np.vstack([tmp["x"] for tmp in self.validate_data_traj])
        num_alpha = len(ALPHA_RANGE)

        # 1. normalize the features by making modal amplitude to 1 for all features
        phi_tilde_scaled = phi_tilde / np.abs(phi_tilde[0, :])
        assert phi_tilde_scaled.shape == phi_tilde.shape

        # 2. augment the complex AX=B problem into an AX=B problem with real entries
        #    since the current package only supports real number arrays

        a = np.hstack([np.real(phi_tilde_scaled), -np.imag(phi_tilde_scaled)])
        b = np.hstack([np.imag(phi_tilde_scaled), np.real(phi_tilde_scaled)])
        phi_tilde_aug = np.vstack([a, b])
        X_aug = np.vstack([X, np.zeros(X.shape)])
        num_data = X.shape[0]
        alphas_enet, coefs_enet_aug, _ = enet_path(
            phi_tilde_aug,
            X_aug,
            l1_ratio=L1_RATIO,
            tol=TOL,
            max_iter=MAX_ITER,
            alphas=ALPHA_RANGE,
            check_input=True,
            verbose=0,
        )
        num_total_eigen_func = int(coefs_enet_aug.shape[1] / 2)

        # get the real and imaginary parts from the complex solution
        coefs_enet_real = coefs_enet_aug[:, :num_total_eigen_func, :]
        coefs_enet_imag = coefs_enet_aug[:, num_total_eigen_func:, :]
        assert coefs_enet_imag.shape == coefs_enet_real.shape

        # combine them into a complex array for final results
        coefs_enet_comp = coefs_enet_real + 1j * coefs_enet_imag

        # 2.5 remove features that are smaller than 'self.truncation_threshold'
        # of the max, because most often
        for i_alpha in range(coefs_enet_comp.shape[2]):
            for i_target in range(coefs_enet_comp.shape[0]):
                coef_cutoff_value = self.truncation_threshold * np.max(
                    abs(coefs_enet_comp[i_target, :, i_alpha])
                )
                index_remove = (
                    abs(coefs_enet_comp[i_target, :, i_alpha]) < coef_cutoff_value
                )
                coefs_enet_comp[i_target, index_remove, i_alpha] = 0 + 0j

        # 2.7 given the selected features, do LS-refit to remove the bias of any kind
        # of regularization
        for i_alpha in range(coefs_enet_comp.shape[2]):
            bool_non_zero = np.linalg.norm(coefs_enet_comp[:, :, i_alpha], axis=0) > 0
            phi_tilde_scaled_reduced = phi_tilde_scaled[:, bool_non_zero]
            coef_enet_comp_reduced_i_alpha = np.linalg.lstsq(
                phi_tilde_scaled_reduced, X
            )[0]
            coefs_enet_comp[
                :, bool_non_zero, i_alpha
            ] = coef_enet_comp_reduced_i_alpha.T
            coefs_enet_comp[:, np.invert(bool_non_zero), i_alpha] = 0

        # 3. compute residual for parameter sweep to draw the trade-off
        coefs_enet = np.abs(coefs_enet_comp)
        residual_array = []
        for i in range(num_alpha):
            residual = np.linalg.norm(
                X
                - np.matmul(phi_tilde_scaled, coefs_enet_comp[:, :, i].T)[:num_data]
                # computed the augmented but only compare the first half rows
            )
            residual /= np.linalg.norm(X)
            residual_array.append(residual)
        residual_array = np.array(residual_array)

        # compute the number of non-zeros
        num_non_zero_all_alpha = []
        num_target_components = coefs_enet_comp.shape[0]
        for ii in range(coefs_enet_comp.shape[2]):
            non_zero_index_per_alpha = []
            for i_component in range(num_target_components):
                non_zero_index_per_alpha_per_target = abs(
                    coefs_enet_comp[i_component, :, ii]
                ) > 0 * np.max(abs(coefs_enet_comp[i_component, :, ii]))
                non_zero_index_per_alpha.append(non_zero_index_per_alpha_per_target)
            non_zero_index_per_alpha_all_targets = np.logical_or.reduce(
                non_zero_index_per_alpha
            )
            num_non_zero_all_alpha.append(np.sum(non_zero_index_per_alpha_all_targets))
        num_non_zero_all_alpha = np.array(num_non_zero_all_alpha)

        # print a table for non-zero alpha
        sparse_error_table = PrettyTable()
        sparse_error_table.field_names = [
            "Index",
            "Alpha",
            "# Non-zero",
            "Reconstruction Error",
        ]
        for i in range(len(ALPHA_RANGE)):
            sparse_error_table.add_row(
                [
                    i,
                    ALPHA_RANGE[i],
                    num_non_zero_all_alpha[i],
                    residual_array[i],
                ]
            )
        print(sparse_error_table)

        # plot figures
        num_target_components = coefs_enet_comp.shape[0]
        alphas_enet_log_negative = -np.log10(alphas_enet)
        top_k_modes_list = np.arange(L)

        # figure set 1 -- sparsity of Koopman mode in reconstructing each target
        if plot:
            for i_component in range(num_target_components):
                plt.figure(figsize=(6, 6))
                for i in top_k_modes_list:
                    i_s = self.small_to_large_error_eigen_index[i]
                    plt.plot(
                        alphas_enet_log_negative,
                        abs(coefs_enet[i_component, i, :]),
                        "-*",
                        label=r"$\lambda_{}$ = {:.2f}".format(
                            i_s, self.eigenvalues_discrete[i_s]
                        ),
                    )
                max_val = np.max(abs(coefs_enet[i_component, :, -1]))
                min_val = np.min(abs(coefs_enet[i_component, :, -1]))
                diss = (max_val - min_val) / 2
                mean = (max_val + min_val) / 2
                plt.xlabel(r"-$\log_{10}(\alpha)$", fontsize=16)
                plt.ylabel(
                    "Absolute Value of Coefficients",
                    fontsize=16,
                )
                plt.ylim([mean - diss * 3, mean + diss * 3])
                plt.title(r"$x_{}$".format(i_component + 1))
                plt.legend(loc="best")
                if save_figure:
                    plt.savefig(
                        self.dir
                        + "multi-elastic-net-coef-"
                        + str(i_component + 1)
                        + ".png",
                        bbox_inches="tight",
                    )
                    plt.close()
                else:
                    plt.tight_layout()
                    plt.show()

            # figure set 2 -- reconstruction MSE vs alpha
            fig = plt.figure(figsize=(6, 6))
            ax1 = fig.add_subplot(111)
            ax2 = ax1.twinx()
            ax1.plot(alphas_enet_log_negative, residual_array, "k*-")
            ax1.set_xlabel(r"-$\log_{10}(\alpha)$", fontsize=16)
            ax1.set_ylabel(
                "Normalized Reconstruction MSE",
                color="k",
                fontsize=16,
            )

            ax2.plot(alphas_enet_log_negative, num_non_zero_all_alpha, "r*-")
            ax2.set_ylabel(
                "Number of Selected Koopman V",
                color="r",
                fontsize=16,
            )

            if save_figure:
                plt.savefig(
                    self.dir + "/multi-elastic-net-mse.png", bbox_inches="tight"
                )
                plt.close()
            else:
                plt.tight_layout()
                plt.show()

        # 4. find the selected index within the top L best eigenmodes for each alpha
        sweep_index_list = []
        for ii, alpha in enumerate(alphas_enet):
            non_zero_index_bool_array = (
                np.linalg.norm(coefs_enet_comp[:, :, ii], axis=0) > 0
            )
            sweep_index_list.append(non_zero_index_bool_array)
        self.sweep_index_list = sweep_index_list

    def prune_model(self, i_alpha, x_train, dt=1):
        """Prune the `pykoopman.koopman.Koopman` model.

        Returns a pruned model that contains most of the functionality of
        the original model.

        Parameters:
            i_alpha (int): The chosen index from the result of sparse linear regression
                to prune the model.
            x_train (numpy.ndarray): The training data, but only the `x`. Used to refit
                the Koopman V since the Koopman eigenmodes are sparsified.
            dt (float, optional): Time step used in the original model. Defaults to 1.

        Returns:
            pruned_model (PrunedKoopman): The pruned model with fewer Koopman V, but
                similar accuracy.
        """
        sweep_bool_index = self.sweep_index_list[i_alpha]
        sweep_index = self.small_to_large_error_eigen_index[: self.L][sweep_bool_index]

        pruned_model = PrunedKoopman(self.model, sweep_index, dt)
        pruned_model = pruned_model.fit(x_train)
        return pruned_model



"""module for implementing a neural network DMD"""
from __future__ import annotations

import pickle
from abc import abstractmethod
from warnings import warn

import lightning as L
import numpy as np
import torch
from pykoopman.regression._base import BaseRegressor
from sklearn.utils.validation import check_is_fitted
from torch import nn
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import DataLoader
from torch.utils.data import Dataset


# todo: add the control version


class MaskedMSELoss(nn.Module):
    """
    Calculates the mean squared error (MSE) loss between `output` and `target`, with
    masking based on `target_lens`. The `max_look_forward` will determine the

    Args:
        max_look_forward

    Returns:
        The MSE loss as a scalar tensor.
    """

    def __init__(self, max_look_forward):
        super().__init__()
        self.max_look_forward = torch.tensor(max_look_forward, dtype=torch.int)

    def forward(self, output, target, target_lens):
        """
        Calculates the MSE loss between `output` and `target`, with masking based on
        `target_lens`.

        Args:
            output (torch.Tensor): The output tensor of shape (batch_size,
            sequence_length, features).
            target (torch.Tensor): The target tensor of shape (batch_size,
            sequence_length, features).
            target_lens (torch.Tensor): A tensor of shape (batch_size,) containing the
            sequence lengths for each item in the batch.

        Returns:
            The MSE loss as a scalar tensor.
        """

        # if target is shorter than output, just cut output off
        if target.size(1) < self.max_look_forward:
            output = output[:, : target.size(1), :]

        # Create mask using target_lens
        mask = torch.zeros_like(output, dtype=torch.bool)
        for i, length in enumerate(target_lens):
            if length > self.max_look_forward:
                length_used = self.max_look_forward
            else:
                length_used = length
            mask[i, :length_used, :] = 1

        # Compute squared differences and apply mask
        squared_diff = torch.pow(output - target, 2)
        squared_diff_masked = torch.where(
            mask, squared_diff, torch.zeros_like(squared_diff)
        )

        # Compute the MSE loss
        mse_loss = squared_diff_masked.sum() / mask.sum()

        return mse_loss


class FFNN(nn.Module):
    """A feedforward neural network with customizable architecture and activation
    functions.

    Args:
        input_size (int): The size of the input layer.
        hidden_sizes (list): A list of the sizes of the hidden layers.
        output_size (int): The size of the output layer.
        activations (str): A string for activation functions for every layer.

    Attributes:
        layers (nn.ModuleList): A list of the neural network layers.
    """

    def __init__(
        self, input_size, hidden_sizes, output_size, activations, include_state=False
    ):
        super(FFNN, self).__init__()

        activations_dict = {
            "relu": nn.ReLU(),
            "sigmoid": nn.Sigmoid(),
            "tanh": nn.Tanh(),
            "swish": nn.SiLU(),
            "elu": nn.ELU(),
            "mish": nn.Mish(),
            "linear": nn.Identity(),
        }
        # whether to directly encode state in observables
        self.include_state = include_state
        if self.include_state:
            output_size_ = output_size - input_size
        else:
            output_size_ = output_size

        # Define the activation
        act = activations_dict[activations]

        # Define the input layer
        self.layers = nn.ModuleList()

        # if linear layer, remove bias
        if activations == "linear":
            bias = False
        else:
            bias = True

        if len(hidden_sizes) == 0:
            # if no hidden layer, then entire NN is just a linear one
            self.layers.append(nn.Linear(input_size, output_size_, bias))
        else:
            self.layers.append(nn.Linear(input_size, hidden_sizes[0], bias))
            if activations != "linear":
                self.layers.append(act)

            # Define the hidden layers
            for i in range(1, len(hidden_sizes)):
                self.layers.append(
                    nn.Linear(hidden_sizes[i - 1], hidden_sizes[i], bias)
                )
                if activations != "linear":
                    self.layers.append(act)

            # Define the last output layer
            self.layers.append(nn.Linear(hidden_sizes[-1], output_size_, bias=False))

    def forward(self, x):
        """Performs a forward pass through the neural network.

        Args:
            x (torch.Tensor): The input tensor to the neural network.

        Returns:
            torch.Tensor: The output tensor of the neural network.
        """
        in_x = x
        for layer in self.layers:
            x = layer(x)

        if self.include_state:
            x = torch.cat((in_x, x), 1)
        return x


class HardCodedLinearLayer(nn.Module):
    def __init__(self, input_size, output_size):

        pass

    def forward(self, x):
        pass


class BaseKoopmanOperator(nn.Module):
    """Base class for Koopman operator models.

    Args:
        dim (int): The dimension of the state space.
        dt (float, optional): The time step size. Defaults to 1.0.
        init_std (float, optional): The standard deviation of the initializer.
            Defaults to 0.1.

    Attributes:
        dim (int): The dimension of the state space.
        dt (torch.Tensor): The time step size.
        init_std (float): The standard deviation of the initializer.

    Note:
        rule for self.init_std: a number between 0.1 and 10 over dt

    """

    def __init__(
        self,
        dim: int,
        dt: float = 1.0,
        init_std: float = 0.1,
    ):
        """
        Initializes the `BaseKoopmanOperator` instance.
        """
        super().__init__()
        self.dim = dim
        self.register_buffer("dt", torch.tensor(dt))
        self.init_std = init_std

    def forward(self, x):
        """
        Computes the forward pass of the `BaseKoopmanOperator`.

        Given `x` as a row vector, return `x @ K.T`

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The output tensor.
        """
        koopman_operator = self.get_discrete_time_Koopman_Operator()
        xnext = torch.matmul(x, koopman_operator.t())  # following pytorch convention
        return xnext

    def get_discrete_time_Koopman_Operator(self):
        """
        Computes the discrete-time Koopman operator.

        Returns:
            torch.Tensor: The discrete-time Koopman operator.
        """
        return torch.matrix_exp(self.dt * self.get_K())

    @abstractmethod
    def get_K(self):
        """
        Computes the matrix K.

        Returns:
            torch.Tensor: The matrix K.
        """
        pass


class StandardKoopmanOperator(BaseKoopmanOperator):
    """
    Standard Koopman operator that only has a diagonal matrix for the Koopman operator.
    """

    def __init__(self, **kwargs):
        """
        Initializes the StandardKoopmanOperator.

        Args:
            **kwargs: Additional keyword arguments.
        """
        super().__init__(**kwargs)
        self.register_parameter(
            "K",
            torch.nn.Parameter(
                torch.nn.init.trunc_normal_(
                    torch.zeros(self.dim, self.dim), std=self.init_std
                )
            ),
        )

    def get_K(self):
        """
        Computes the Koopman operator.

        Returns:
            The Koopman operator.
        """
        return self.K


class HamiltonianKoopmanOperator(BaseKoopmanOperator):
    """
    Hamiltonian Koopman operator that has an off-diagonal matrix for the Koopman
    operator.
    """

    def __init__(self, **kwargs):
        """
        Initializes the HamiltonianKoopmanOperator.

        Args:
            **kwargs: Additional keyword arguments.
        """
        super().__init__(**kwargs)
        self.register_parameter(
            "off_diagonal",
            torch.nn.Parameter(
                torch.nn.init.trunc_normal_(
                    torch.zeros(self.dim, self.dim), std=self.init_std
                )
            ),
        )

    def get_K(self):
        """
        Computes the Koopman operator.

        Returns:
            The Koopman operator.
        """
        return self.off_diagonal - self.off_diagonal.t()


class DissipativeKoopmanOperator(BaseKoopmanOperator):
    """
    Dissipative Koopman operator that has an off-diagonal and a diagonal matrix for the
    Koopman operator.
    """

    def __init__(self, **kwargs):
        """
        Initializes the DissipativeKoopmanOperator.

        Args:
            **kwargs: Additional keyword arguments.
        """
        super().__init__(**kwargs)
        self.register_parameter(
            "off_diagonal",
            torch.nn.Parameter(
                torch.nn.init.trunc_normal_(
                    torch.zeros(self.dim, self.dim), std=self.init_std
                )
            ),
        )
        self.register_parameter(
            "diagonal",
            torch.nn.Parameter(
                -torch.pow(
                    torch.nn.init.trunc_normal_(
                        torch.zeros(self.dim), std=self.init_std
                    ),
                    2,
                )
            ),
        )

    def get_K(self):
        """
        Computes the Koopman operator.

        Returns:
            The Koopman operator.
        """
        return torch.diag(self.diagonal) + self.off_diagonal - self.off_diagonal.t()


class DLKoopmanRegressor(L.LightningModule):
    """
    Deep Learning Koopman Regressor module using a Feedforward Neural Network
    encoder and decoder to learn the Koopman operator for a given dynamical system.

    Args:
        mode (str): Type of Koopman operator to use - "Standard", "Hamiltonian" or
            "Dissipative". Defaults to None.
        dt (float): Time step of the Koopman operator. Defaults to 1.0.
        look_forward (int): Number of time steps to predict in the future.
            Defaults to 1.
        config_encoder (dict): Dictionary containing encoder configurations
            - input_size, output_size, hidden_sizes, activations. Defaults to {}.
        config_decoder (dict): Dictionary containing decoder configurations
            - input_size, output_size, hidden_sizes, activations. Defaults to {}.
        lbfgs (bool): Use L-BFGS optimizer. Defaults to False.

    Attributes:
        input_size (int): Size of input to the encoder.
        output_size (int): Size of output from the encoder.
        _encoder (FFNN): Feedforward Neural Network encoder.
        _decoder (FFNN): Feedforward Neural Network decoder.
        _koopman_propagator (BaseKoopmanOperator): Type of Koopman operator used.
        look_forward (int): Number of time steps to predict in the future.
        using_lbfgs (bool): Use L-BFGS optimizer.
        masked_loss_metric (MaskedMSELoss): Mean Squared Error Loss function.
    """

    def __init__(
        self,
        mode=None,
        dt=1.0,
        look_forward=1,
        config_encoder=dict(),
        config_decoder=dict(),
        lbfgs=False,
        std_koopman=1e-1,
        include_state=False,
    ):
        super(DLKoopmanRegressor, self).__init__()

        self.input_size = config_encoder["input_size"]
        self.output_size = config_encoder["output_size"]

        self._encoder = FFNN(
            input_size=config_encoder["input_size"],
            hidden_sizes=config_encoder["hidden_sizes"],
            output_size=config_encoder["output_size"],
            activations=config_encoder["activations"],
            include_state=include_state,
        )

        self._decoder = FFNN(
            input_size=config_decoder["input_size"],
            hidden_sizes=config_decoder["hidden_sizes"],
            output_size=config_decoder["output_size"],
            activations=config_decoder["activations"],
        )

        if mode == "Dissipative":
            self._koopman_propagator = DissipativeKoopmanOperator(
                dim=config_encoder["output_size"], dt=dt, init_std=std_koopman
            )
        elif mode == "Hamiltonian":
            self._koopman_propagator = HamiltonianKoopmanOperator(
                dim=config_encoder["output_size"], dt=dt, init_std=std_koopman
            )
        else:
            self._koopman_propagator = StandardKoopmanOperator(
                dim=config_encoder["output_size"], dt=dt, init_std=std_koopman
            )

        self.look_forward = look_forward
        self.using_lbfgs = lbfgs

        # self.masked_loss_metric = MaskedMSELoss(1)
        self.masked_loss_metric = MaskedMSELoss(self.look_forward)

        if self.using_lbfgs:
            self.automatic_optimization = False

            def training_step(batch, batch_idx):
                optimizer = self.optimizers()

                def closure():

                    # unpack batch data
                    x, y, ys = batch

                    # get the max look forward in this batch
                    batch_look_forward = ys.max()

                    # encode x
                    encoded_x = self._encoder(x)

                    # future unroll look_forward
                    phi_seq = self._propagate_encoded_n_steps(
                        encoded_x, n=batch_look_forward
                    )

                    # standard RNN loss
                    decoded_y_seq_rnn = torch.zeros(
                        (x.size(0), self.look_forward, self.input_size),
                        device=self.device,
                    )

                    for i in range(batch_look_forward):
                        decoded_y_seq_rnn[:, i, :] = self._decoder(phi_seq[:, i, :])
                    rnn_loss = self.masked_loss_metric(decoded_y_seq_rnn, y, ys)

                    # autoencoder reconstruction loss
                    # for x
                    decoded_x = self._decoder(encoded_x)
                    rec_loss = torch.nn.functional.mse_loss(decoded_x, x)

                    # for y_seq
                    decoded_y_seq_rec = torch.zeros(
                        (x.size(0), self.look_forward, self.input_size),
                        device=self.device,
                    )
                    for i in range(batch_look_forward):
                        decoded_y_seq_rec[:, i, :] = self._decoder(
                            self._encoder(y[:, i, :])
                        )
                    rec_loss += self.masked_loss_metric(decoded_y_seq_rec, y, ys)

                    loss = rnn_loss + rec_loss

                    optimizer.zero_grad()
                    self.manual_backward(loss)

                    self.log("loss", loss, prog_bar=True)
                    self.log("rec_loss", rec_loss, prog_bar=True)
                    self.log("rnn_loss", rnn_loss, prog_bar=True)

                    return loss

                optimizer.step(closure=closure)

            self.training_step = training_step

        self.save_hyperparameters()

    def forward(self, x, n=1):
        """
        Propagates input tensor through the model to obtain predicted output tensor
        after n steps.

        Args:
            x: Input tensor with shape (batch_size, input_size).
            n (int): Number of steps to propagate.

        Returns:
            decoded: Output tensor with shape (batch_size, output_size).

        """
        encoded = self._encoder(x)
        phi_seq = self._propagate_encoded_n_steps(encoded, n)
        decoded = self._decoder(phi_seq[:, -1, :])
        return decoded

    def forward_all(self, x, n):
        """
        Forward pass of the Koopman Regressor for a given sequence of input states `x`.
        This method returns the decoded sequence for all steps within the horizon `n`.

        Args:
            x (torch.Tensor): The input state sequence with shape `(batch_size, seq_len,
                input_size)`.
            n (int): The maximum horizon for which to generate the output sequence.

        Returns:
            decoded (torch.Tensor): The decoded sequence with shape `(batch_size, n,
                input_size)`.
        """
        encoded = self._encoder(x)
        phi_seq = self._propagate_encoded_n_steps(encoded, n)
        decoded = torch.zeros(x.size(0), n, self.input_size)
        for i in range(n):
            decoded[:, i, :] = self._decoder(phi_seq[:, i, :])
        return decoded

    def _propagate_encoded_n_steps(self, encoded, n):
        """
        Propagates the encoded tensor linearly in the encoded space for n steps.

        Args:
            encoded (torch.Tensor): The encoded tensor of shape (batch_size,
            encoded_size). n (int): The number of steps to propagate.

        Returns:
            torch.Tensor: The propagated encoded tensor of shape (batch_size, n,
            encoded_size).
        """
        encoded_future = []
        for i in range(n):
            encoded = self._koopman_propagator(encoded)
            encoded_future.append(encoded)
        return torch.stack(encoded_future, 1)

    def training_step(self, batch, batch_idx):
        """
        Defines a training step for the DL Koopman Regressor.

        Args:
            batch: tuple of (x, y, ys), representing the input data,
                the true output data, and the sequence length for
                each sample in the batch.
            batch_idx: integer, the index of the batch.

        Returns:
            tensor representing the loss value for this training step.
        """
        # unpack batch data
        x, y, ys = batch

        # get the max look forward in this batch
        batch_look_forward = ys.max()

        # encode x
        encoded_x = self._encoder(x)

        # future unroll look_forward
        phi_seq = self._propagate_encoded_n_steps(encoded_x, n=batch_look_forward)

        # standard RNN loss
        decoded_y_seq_rnn = torch.zeros(
            (x.size(0), self.look_forward, self.input_size), device=self.device
        )

        for i in range(batch_look_forward):
            decoded_y_seq_rnn[:, i, :] = self._decoder(phi_seq[:, i, :])
        rnn_loss = self.masked_loss_metric(decoded_y_seq_rnn, y, ys)

        # autoencoder reconstruction loss
        # for x
        decoded_x = self._decoder(encoded_x)
        rec_loss = torch.nn.functional.mse_loss(decoded_x, x)

        # for y_seq
        decoded_y_seq_rec = torch.zeros(
            (x.size(0), self.look_forward, self.input_size), device=self.device
        )
        for i in range(batch_look_forward):
            decoded_y_seq_rec[:, i, :] = self._decoder(self._encoder(y[:, i, :]))
        rec_loss += self.masked_loss_metric(decoded_y_seq_rec, y, ys)

        loss = rnn_loss + rec_loss

        self.log("loss", loss, prog_bar=True)
        self.log("rec_loss", rec_loss, prog_bar=True)
        self.log("rnn_loss", rnn_loss, prog_bar=True)
        return loss

    def configure_optimizers(self):
        """Configures and returns the optimizer to use for training.

        If using LBFGS optimizer, set `using_lbfgs` attribute to True when
        initializing the DLKoopmanRegressor instance.

        Returns:
            An instance of torch.optim.Optimizer to use for training.
        """
        if self.using_lbfgs:
            optimizer = torch.optim.LBFGS(
                self.parameters(),
                lr=1,
                history_size=100,
                max_iter=20,
                line_search_fn="strong_wolfe",
            )
        else:
            optimizer = torch.optim.Adam(self.parameters(), lr=1e-3)
        return optimizer


class SeqDataDataset(Dataset):
    """
    A PyTorch Dataset class to handle sequential data in the format of (x, y, ys),
    where x is the input sequence, y is the target output sequence and ys is a vector
    indicating the maximum look-ahead distance.

    Args:
        x (torch.Tensor): The input sequence tensor of shape (batch_size,
            sequence_length, input_size).
        y (torch.Tensor): The output sequence tensor of shape (batch_size,
            sequence_length, output_size).
        ys (torch.Tensor): The maximum look-ahead distance tensor of shape
            (batch_size,).
        transform (callable, optional): Optional normalization function to apply to
            x and y.

    Returns:
        torch.Tensor: The preprocessed input sequence tensor.
        torch.Tensor: The preprocessed target output sequence tensor.
        torch.Tensor: The maximum look-ahead distance tensor.
    """

    def __init__(self, x, y, ys, transform=None):
        self.x = x.squeeze(1)
        self.y = y
        self.ys = ys
        self.normalization = transform

    def __len__(self):
        return len(self.ys)

    def __getitem__(self, idx):
        x = self.x[idx].clone()
        y = self.y[idx].clone()
        ys = self.ys[idx].clone()

        if self.normalization:
            x = self.normalization(x)
            y = self.normalization(y)

        return x, y, ys


class TensorNormalize(nn.Module):
    """
    Normalizes the input tensor by subtracting the mean and dividing by the standard
    deviation.

    Args:
        mean (float or tensor): The mean value to be subtracted from the input tensor.
        std (float or tensor): The standard deviation value to divide the input tensor
            by.
    """

    def __init__(self, mean, std):
        super().__init__()
        self.mean = mean
        self.std = std

    def forward(self, tensor: torch.Tensor):
        """
        Forward pass of the normalization module.

        Args:
            tensor (tensor): The input tensor to be normalized.

        Returns:
            The normalized tensor.
        """
        return torch.divide((tensor - self.mean), self.std)
        # return # tensor.copy_(tensor.sub_(self.mean).div_(self.std))

    def __repr__(self) -> str:
        """
        Returns a string representation of the TensorNormalize module.

        Returns:
            A string representation of the module.
        """
        return f"{self.__class__.__name__}(mean={self.mean}, std={self.std})"


class InverseTensorNormalize(nn.Module):
    """
    A PyTorch module that performs inverse normalization on input tensors using
    a given mean and standard deviation.

    Args:
        mean (float or sequence): The mean used for normalization.
        std (float or sequence): The standard deviation used for normalization.

    Example:
        >>> mean = [0.5, 0.5, 0.5]
        >>> std = [0.5, 0.5, 0.5]
        >>> inv_norm = InverseTensorNormalize(mean, std)
        >>> normalized_tensor = torch.tensor([[-1.0, 0.0, 1.0], [-0.5, 0.0, 0.5]])
        >>> output = inv_norm(normalized_tensor)

    Attributes:
        mean (float or sequence): The mean used for normalization.
        std (float or sequence): The standard deviation used for normalization.
    """

    def __init__(self, mean, std):
        super().__init__()
        self.mean = mean
        self.std = std

    def forward(self, tensor: torch.Tensor):
        return torch.multiply(tensor, self.std) + self.mean
        # return tensor.copy_(tensor.mul_(self.std).add_(self.mean))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(mean={self.mean}, std={self.std})"


class SeqDataModule(L.LightningDataModule):
    """
    Class for creating sequence data dataloader for training and validation.

    Args:
        data_tr: List of 2D numpy.ndarray representing training data trajectories.
        data_val: List of 2D numpy.ndarray representing validation data trajectories.
            Can be None.
        look_forward: Number of time steps to predict forward.
        batch_size: Size of each batch of data.
        normalize: Whether to normalize the input data or not. Default is True.
        normalize_mode: The type of normalization to use. Either "equal" or "max".
            Default is "equal".
        normalize_std_factor: Scaling factor for standard deviation during
            normalization. Default is 2.0.

    Methods:
        prepare_data(): Prepares the data by converting to time-delayed data and
            computing mean and std if normalize is True.
        setup(stage=None): Sets up training and validation datasets.
        train_dataloader(): Returns a DataLoader for training data.
        val_dataloader(): Returns a DataLoader for validation data.
        convert_seq_list_to_delayed_data(data_list, look_back, look_forward): Converts
            list of sequences to time-delayed data.
        collate_fn(batch): Custom collate function to be used with DataLoader.

    Returns:
        A SeqDataModule object.
    """

    def __init__(
        self,
        data_tr,
        data_val,
        look_forward=10,
        batch_size=32,
        normalize=True,
        normalize_mode="equal",
        normalize_std_factor=2.0,
    ):
        """
        Initialize a SeqDataModule.

        Args:
            data_tr (Union[str, List[np.ndarray]]): Training data. Can be either a
                list of 2D numpy arrays, each 2D numpy array representing a trajectory,
                or the path to a pickle file containing such a list.
            data_val (Optional[Union[str, List[np.ndarray]]]): Validation data.
                Can be either a list of 2D numpy arrays, each 2D numpy array
                    representing a trajectory, or the path to a pickle file
                    containing such a list.
            look_forward (int): Number of time steps to predict into the future.
            batch_size (int): Number of samples per batch.
            normalize (bool): Whether to normalize the data. Default is True.
            normalize_mode (str): Mode for normalization. Can be either "equal"
                or "max". "equal" divides by the standard deviation, while "max"
                divides by the maximum absolute value of the data. Default is "equal".
            normalize_std_factor (float): Scaling factor for the standard deviation in
                normalization. Default is 2.0.

        Returns:
            None.
        """
        super().__init__()
        # input data_tr or data_val is a list of 2D np.ndarray. each 2d
        # np.ndarray is a trajectory, and the axis 0 is number of samples, axis 1 is
        # the number of system state
        self.data_tr = data_tr
        self.data_val = data_val
        self.look_forward = look_forward
        self.batch_size = batch_size
        self.look_back = 1
        self.normalize = normalize
        self.normalize_mode = normalize_mode
        self.normalization = None
        self.inverse_transform = None
        self.normalize_std_factor = normalize_std_factor

    def prepare_data(self):
        """
        Preprocesses the input training and validation data by checking their types,
        checking for normalization, finding the mean and standard deviation of
        the training data (if normalization is enabled), and creating time-delayed data
        from the input data.

        Raises:
            ValueError: If the training data is None or has an invalid type.
            ValueError: If the validation data has an invalid type.
            TypeError: If the data is complex or not float.

        """
        # train data
        if self.data_tr is None:
            raise ValueError("You must feed training data!")
        if isinstance(self.data_tr, list):
            data_list = self.data_tr
        elif isinstance(self.data_tr, str):
            f = open(self.data_tr, "rb")
            data_list = pickle.load(f)
        else:
            raise ValueError("Wrong type of `self.data_tr`")

        # check train data
        data_list = self.check_list_of_nparray(data_list)

        # find the mean, std
        if self.normalize:
            stacked_data_list = np.vstack(data_list)
            mean = stacked_data_list.mean(axis=0)
            std = stacked_data_list.std(axis=0)

            # zero mean so easier for downstream
            self.mean = torch.FloatTensor(mean) * 0
            # default = 2.0, more stable
            self.std = torch.FloatTensor(std) * self.normalize_std_factor

            if self.normalize_mode == "max":
                self.std = torch.ones_like(self.std) * self.std.max()

            # prevent divide by zero error
            for i in range(len(self.std)):
                if self.std[i] < 1e-6:
                    self.std[i] += 1e-3

            # get transform
            self.normalization = TensorNormalize(self.mean, self.std)

            # get inverse transform
            self.inverse_transform = InverseTensorNormalize(self.mean, self.std)

        # create time-delayed data
        self._tr_x, self._tr_yseq, self._tr_ys = self.convert_seq_list_to_delayed_data(
            data_list, self.look_back, self.look_forward
        )

        # validation data
        if self.data_val is not None:
            # raise ValueError("You need to feed validation data!")
            if isinstance(self.data_val, list):
                data_list = self.data_val
            elif isinstance(self.data_val, str):
                f = open(self.data_val, "rb")
                data_list = pickle.load(f)
            else:
                raise ValueError("Wrong type of `self.data_val`")

            # check val data
            data_list = self.check_list_of_nparray(data_list)

            # create time-delayed data
            (
                self._val_x,
                self._val_yseq,
                self._val_ys,
            ) = self.convert_seq_list_to_delayed_data(
                data_list, self.look_back, self.look_forward
            )
        else:
            warn("Warning: no validation data prepared")

    def setup(self, stage=None):
        """
        Prepares the train and validation datasets for the Lightning module.
        The train dataset is created from the training data specified in the
        constructor by creating time-delayed versions of the input/output sequences.
        If `normalize` is True, the data is normalized using the mean and standard
        deviation of the training data. The validation dataset is created from the
        validation data specified in the constructor in the same way as the training
        dataset. If `normalize` is True, it is also normalized using the mean and
        standard deviation of the training data. If `stage` is not "fit",
        an exception is raised as the `setup()` method has not been implemented
        for other stages.

        Args:
            stage: The stage of training, validation or testing (default is None).

        Raises:
            NotImplementedError: If `stage` is not "fit".
        """
        # Load data and split into train and validation sets here
        # Assign train/val datasets for use in dataloaders
        if stage == "fit":
            self.tr_dataset = SeqDataDataset(
                self._tr_x, self._tr_yseq, self._tr_ys, self.normalization
            )
            if self.data_val is not None:
                self.val_dataset = SeqDataDataset(
                    self._val_x, self._val_yseq, self._val_ys, self.normalization
                )
        else:
            raise NotImplementedError("We didn't implement for stage not `fit`")

    def train_dataloader(self):
        return DataLoader(
            self.tr_dataset, self.batch_size, shuffle=True, collate_fn=self.collate_fn
        )

    def val_dataloader(self):
        return DataLoader(
            self.val_dataset, self.batch_size, shuffle=True, collate_fn=self.collate_fn
        )

    def convert_seq_list_to_delayed_data(self, data_list, look_back, look_forward):
        """
        Converts a list of sequences to time-delayed data by extracting subsequences
        of length `look_back` and `look_forward` from each sequence in the list.

        Args:
            data_list (List[np.ndarray]): A list of 2D numpy arrays. Each array
                represents a trajectory, with axis 0 representing the number of samples
                and axis 1 representing the number of system states.
            look_back (int): The number of previous time steps to include in each
                subsequence.
            look_forward (int): The number of future time steps to include in each
                subsequence.

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor]: A tuple containing three
                tensors:
            1) The time-delayed input data, with shape (num_samples, look_back,
                num_system_states).
            2) The time-delayed output data, with shape (num_samples, look_forward,
                num_system_states).
            3) The sequence lengths of the output data, with shape (num_samples,).
        """
        time_delayed_x_list = []
        time_delayed_yseq_list = []
        for seq in data_list:
            # if self.look_forward + self.look_back > len(seq):
            #     raise ValueError("look_forward too large")
            n_sub_traj = len(seq) - look_back - look_forward + 1
            if n_sub_traj >= 1:
                for i in range(len(seq) - look_back - look_forward + 1):
                    time_delayed_x_list.append(seq[i : i + look_back])
                    time_delayed_yseq_list.append(
                        seq[i + look_back : i + look_back + look_forward]
                    )
            else:
                # only 1 traj, just to predict to its end
                time_delayed_x_list.append(seq[0:1])
                time_delayed_yseq_list.append(seq[1:])
        time_delayed_yseq_lens_list = [x.shape[0] for x in time_delayed_yseq_list]

        # convert data to tensor
        time_delayed_x = torch.FloatTensor(np.array(time_delayed_x_list))
        time_delayed_yseq = pad_sequence(
            [torch.FloatTensor(x) for x in time_delayed_yseq_list], True
        )
        time_delayed_yseq_lens = torch.LongTensor(time_delayed_yseq_lens_list)
        return time_delayed_x, time_delayed_yseq, time_delayed_yseq_lens

    def collate_fn(self, batch):
        """
        Collates a batch of data.

        Args:
            batch: A list of tuples where each tuple represents a sample containing
                the input sequence `x`, the output sequence `y`, and the maximum
                number of steps to predict `ys`.

        Returns:
            A tuple containing the input sequences as a stacked tensor, the output
            sequences as a stacked tensor, and the maximum number of steps to predict
            as a stacked tensor.

        """
        x_batch, y_batch, ys_batch = zip(*batch)
        xx = torch.stack(x_batch, 0)
        yy = torch.stack(y_batch, 0)
        ys = torch.stack(ys_batch, 0)
        return xx, yy, ys

    @classmethod
    def check_list_of_nparray(cls, data_list):
        """
        Check if the input is a list of numpy arrays, and convert data to float32 if
        float64.

        Args:
            data_list (List[np.ndarray]): A list of numpy arrays representing system
                states.

        Returns:
            List[np.ndarray]: The input list of numpy arrays converted to float32.

        Raises:
            TypeError: If the input data is complex or not float.
        """
        # check if data is complex
        if any(np.iscomplexobj(x) for x in data_list):
            raise TypeError("Complex data is not supported")

        # check if data has float64
        if any(x.dtype is np.float64 for x in data_list):
            warn("Found float64 data. Will convert to float32")

        # convert data to float32 if float64
        for i, data_traj in enumerate(data_list):
            if "float" not in data_traj.dtype.name:
                raise TypeError("Found data is not float")
            if data_traj.dtype.name == "float64":
                data_list[i] = data_traj.astype("float32")

        return data_list


class NNDMD(BaseRegressor):
    """Implementation of Nonlinear Dynamic Mode Decomposition (NNDMD).

    Args:
        mode (str): NNDMD mode, `Dissipative` or `Hamiltonian` or else (default: None).
        dt (float): Time step (default: 1.0).
        look_forward (int): Number of steps to look forward (default: 1).
        config_encoder (dict): Configuration for the encoder network
            (default: dict(input_size=2, hidden_sizes=[32]*2, output_size=6,
            activations='tanh')).
        config_decoder (dict): Configuration for the decoder network
            (default: dict(input_size=6, hidden_sizes=[32]*2, output_size=2,
            activations='linear')).
        batch_size (int): Batch size (default: 16).
        lbfgs (bool): Whether to use L-BFGS optimizer (default: False).
        normalize (bool): Whether to normalize data (default: True).
        normalize_mode (str): Normalization mode, `max` or `equal`
            (default: 'equal').
        normalize_std_factor (float): Standard deviation factor for normalization
            (default: 2.0).
        trainer_kwargs (dict): Arguments for the `pytorch_lightning.Trainer`
            (default: {}).

    Attributes:
        coef_ (np.ndarray): Koopman operator coefficients.
        state_matrix_ (np.ndarray): State matrix of the Koopman operator.
        eigenvalues_ (np.ndarray): Eigenvalues of the Koopman operator.
        eigenvectors_ (np.ndarray): Eigenvectors of the Koopman operator.
        ur (np.ndarray): Effective linear transformation.
        unnormalized_modes (np.ndarray): Unnormalized modes.

    Note:
        The `n_samples_` attribute is meaningless for this class.
        The `dt` argument is only included to please the regressor class and has no
            real use.

    """

    def __init__(
        self,
        mode=None,
        dt=1.0,
        look_forward=1,
        config_encoder=dict(
            input_size=2, hidden_sizes=[32] * 2, output_size=6, activations="tanh"
        ),
        config_decoder=dict(
            input_size=6, hidden_sizes=[32] * 2, output_size=2, activations="linear"
        ),
        batch_size=16,
        lbfgs=False,
        normalize=True,
        normalize_mode="equal",
        normalize_std_factor=2.0,
        std_koopman=1e-1,
        include_state=False,
        trainer_kwargs={},
    ):
        """Initializes the NNDMD model."""
        self.mode = mode
        self.look_forward = look_forward
        self.config_encoder = config_encoder
        self.config_decoder = config_decoder
        self.lbfgs = lbfgs
        self.normalize = normalize
        self.normalize_mode = normalize_mode
        self.dt = dt
        self.trainer_kwargs = trainer_kwargs
        self.normalize_std_factor = normalize_std_factor
        self.batch_size = batch_size
        self.std_koopman = std_koopman
        self.include_state = include_state

        # build DLK regressor
        self._regressor = DLKoopmanRegressor(
            mode, dt, look_forward, config_encoder, config_decoder, lbfgs, std_koopman
        )

    def fit(self, x, y=None, dt=None):
        """fit the NNDMD model with data x,y

        Args:
            x (np.ndarray or list): The training input data. If a 2D numpy array,
                then it represents a single time-series  and each row represents a
                state, otherwise it should be a list of 2D numpy arrays.
            y (np.ndarray or list, optional): The target output data,
                corresponding to `x`. If `None`, `x` is assumed to contain the target
                data in its second half. Defaults to `None`.
            dt (float, optional): The time step used to generate `x`.
                Defaults to `None`.

        Returns:
            None. The fitted model is stored in the class attribute `_regressor`.
        """
        # build trainer
        self.trainer = L.Trainer(**self.trainer_kwargs)

        self.n_input_features_ = self.config_encoder["input_size"]

        # create the data module
        # case  1: a single traj, x is 2D np.ndarray, no validation
        if y is None and isinstance(x, np.ndarray) and x.ndim == 2:
            t0, t1 = x[:-1], x[1:]
            list_of_traj = [np.stack((t0[i], t1[i]), 0) for i in range(len(x) - 1)]
            self.dm = SeqDataModule(
                list_of_traj,
                None,
                self.look_forward,
                self.batch_size,
                self.normalize,
                self.normalize_mode,
                self.normalize_std_factor,
            )
            self.n_samples_ = len(list_of_traj)

        # case 2: x, y are 2D np.ndarray, no validation
        elif (
            isinstance(x, np.ndarray)
            and isinstance(y, np.ndarray)
            and x.ndim == 2
            and y.ndim == 2
        ):
            t0, t1 = x, y
            list_of_traj = [np.stack((t0[i], t1[i]), 0) for i in range(len(x) - 1)]
            self.dm = SeqDataModule(
                list_of_traj,
                None,
                self.look_forward,
                self.batch_size,
                self.normalize,
                self.normalize_mode,
                self.normalize_std_factor,
            )
            self.n_samples_ = len(list_of_traj)

        # case 3: only training data, x is a list of trajectories, y is None
        elif isinstance(x, list) and y is None:
            self.dm = SeqDataModule(
                x,
                None,
                self.look_forward,
                self.batch_size,
                self.normalize,
                self.normalize_mode,
                self.normalize_std_factor,
            )
            self.n_samples_ = len(x)

        # case 4: x, y are two lists of trajectories, we have validation data
        elif isinstance(x, list) and isinstance(y, list):
            self.dm = SeqDataModule(
                x,
                y,
                self.look_forward,
                self.batch_size,
                self.normalize,
                self.normalize_mode,
                self.normalize_std_factor,
            )
            self.n_samples_ = len(x)
        else:
            raise ValueError("check `x` and `y` for `self.fit`")

        # trainer starts to train
        self.trainer.fit(self._regressor, self.dm)

        # compute Koopman operator information
        self._state_matrix_ = (
            self._regressor._koopman_propagator.get_discrete_time_Koopman_Operator()
            .detach()
            .numpy()
        )
        [self._eigenvalues_, self._eigenvectors_] = np.linalg.eig(self._state_matrix_)

        self._coef_ = self._state_matrix_

        # obtain effective linear transformation
        decoder_weight_list = []
        for i in range(len(self._regressor._decoder.layers)):
            decoder_weight_list.append(
                self._regressor._decoder.layers[i].weight.detach().numpy()
            )
        if len(decoder_weight_list) > 1:
            self._ur = np.linalg.multi_dot(decoder_weight_list[::-1])
        else:
            self._ur = decoder_weight_list[0]

        if self.normalize:
            std = self.dm.inverse_transform.std
            self._ur = np.diag(std) @ self._ur

        self._unnormalized_modes = self._ur @ self._eigenvectors_

    def predict(self, x, n=1):
        """
        Predict the system state after n steps away from x_0 = x.

        Args:
            x (numpy.ndarray or torch.Tensor): Input data of shape
                (n_samples, n_features).
            n (int): Number of steps to predict the system state into the future.

        Returns:
            numpy.ndarray: Predicted system state after n steps, of shape
                (n_samples, n_features).

        Note:
            By default, the model is stored on the CPU for inference.
        """
        self._regressor.eval()
        x = self._convert_input_ndarray_to_tensor(x)

        with torch.no_grad():
            # print("inference device = ", self._regressor.device)

            if self.normalize:
                y = self.dm.normalization(x)
                y = self._regressor(y, n)
                y = self.dm.inverse_transform(y).numpy()
            else:
                y = self._regressor(x, n).numpy()
            return y

    def simulate(self, x, n_steps):
        """
        Simulate the system forward in time for `n_steps` steps starting from `x`.

        Args:
            x (np.ndarray or torch.Tensor): The initial state of the system.
            Should be a 2D array/tensor.
            n_steps (int): The number of time steps to simulate the system forward.

        Returns:
            np.ndarray: The simulated states of the system. Will be of shape
            `(n_steps+1, n_features)`.
        """
        self._regressor.eval()
        x = self._convert_input_ndarray_to_tensor(x)
        x_future = torch.zeros([n_steps + 1, x.size(1)])
        x_future[0] = x
        with torch.no_grad():
            for i in range(n_steps):
                if self.normalize:
                    y = self.dm.normalization(x)
                    y = self._regressor(y, i + 1)
                    x_future[i + 1] = self.dm.inverse_transform(y)
                else:
                    x_future[i + 1] = self._regressor(x, i + 1)

            return x_future.numpy()

    @property
    def A(self):
        """Returns the state transition matrix A of the NNDMD model.

        Returns
        -------
        A : numpy.ndarray
            The state transition matrix of shape (n_states, n_states), where
            n_states is the number of states in the model.
        """
        return self._state_matrix_

    @property
    def B(self):
        # todo: we don't have control considered in nndmd for now
        pass

    @property
    def C(self):
        """
        Returns the matrix C representing the effective linear transformation
        from the observables to the Koopman modes. The matrix C is computed during
        the fit process as the product of the decoder weights of the trained
        autoencoder network.

        Returns:
        --------
        numpy.ndarray of shape (n_koopman, n_features)
            The matrix C.
        """
        return self._ur

    @property
    def W(self):
        """
        Returns the matrix W representing the Koopman modes. The matrix W is computed
        during the fit process as the eigenvectors of the Koopman operator.

        Returns:
        --------
        numpy.ndarray of shape (n_koopman, n_koopman)
            The matrix V, where each column represents a Koopman mode.
        """
        return self._unnormalized_modes

    def phi(self, x_col):
        return self._compute_phi(x_col)

    def psi(self, x_col):
        return self._compute_psi(x_col)

    def _compute_phi(self, x_col):
        """
        Computes the Koopman observable vector `phi(x)` for input `x`.

        Args:
            x (np.ndarray or torch.Tensor): The input state vector or tensor.

        Returns:
            phi (np.ndarray): The Koopman observable vector `phi(x)` for input `x`.
        """
        if x_col.ndim == 1:
            x_col = x_col.reshape(-1, 1)
        x = x_col.T

        self._regressor.eval()
        x = self._convert_input_ndarray_to_tensor(x)

        if self.normalize:
            x = self.dm.normalization(x)
        phi = self._regressor._encoder(x).detach().numpy().T
        return phi

    def _compute_psi(self, x_col):
        """
        Computes the Koopman eigenfunction expansion coefficients `psi(x)` given `x`.

        Args:
            x (numpy.ndarray): Input data of shape `(n_samples, n_features)`.

        Returns:
            numpy.ndarray: Koopman eigenfunction expansion coefficients `psi(x)`
                of shape `(n_koopman, n_samples)`.
        """
        if x_col.ndim == 1:
            x_col = x_col.reshape(-1, 1)
        # x = x_col.T

        phi = self._compute_phi(x_col)
        psi = np.linalg.inv(self._eigenvectors_) @ phi
        return psi

    def _convert_input_ndarray_to_tensor(self, x):
        """
        Converts input numpy ndarray to PyTorch tensor with appropriate dtype and
        device.

        Args:
            x (np.ndarray or torch.Tensor): Input data as numpy ndarray or PyTorch
                tensor.

        Returns:
            torch.Tensor: Input data as PyTorch tensor.

        Raises:
            TypeError: If input data is not a numpy ndarray or PyTorch tensor.
            ValueError: If input array has more than 2 dimensions.
        """
        if isinstance(x, np.ndarray):
            if x.ndim > 2:
                raise ValueError("input array should be 1 or 2D")
            if x.ndim == 1:
                x = x.reshape(1, -1)
            # convert to a float32
            # if x.dtype == np.float64:
            x = torch.FloatTensor(x)
        elif isinstance(x, torch.Tensor):
            if x.ndim != 2:
                raise ValueError("input tensor `x` must be a 2d tensor")
        return x

    @property
    def coef_(self):
        check_is_fitted(self, "_coef_")
        return self._coef_

    @property
    def state_matrix_(self):
        return self._state_matrix_

    @property
    def eigenvalues_(self):
        check_is_fitted(self, "_eigenvalues_")
        return self._eigenvalues_

    @property
    def eigenvectors_(self):
        check_is_fitted(self, "_eigenvectors_")
        return self._eigenvectors_

    @property
    def unnormalized_modes(self):
        check_is_fitted(self, "_unnormalized_modes")
        return self._unnormalized_modes

    @property
    def ur(self):
        check_is_fitted(self, "_ur")
        return self._ur


if __name__ == "__main__":
    pass



"""module for Linear observables"""
from __future__ import annotations

import numpy as np
from sklearn.utils.validation import check_is_fitted

from ..common import validate_input
from ._base import BaseObservables


class Identity(BaseObservables):
    """
    A dummy observables class that simply returns its input.
    """

    def __init__(self):
        """
        Initialize the Identity class.

        This constructor initializes the Identity class which simply returns its input
        when transformed.
        """
        super().__init__()
        self.include_state = True

    def fit(self, x, y=None):
        """
        Fit the model to the provided measurement data.

        Args:
            x (array-like): The measurement data to be fit. It must have a shape of
                (n_samples, n_input_features).
            y (None): This parameter is retained for sklearn compatibility.

        Returns:
            self: Returns a fit instance of the class `pykoopman.observables.Identity`.

        Note:
            only identity mapping is supported for list of arb trajectories
        """
        x = validate_input(x)
        if not isinstance(x, list):
            self.n_input_features_ = self.n_output_features_ = x.shape[1]
            self.measurement_matrix_ = np.eye(x.shape[1]).T
        else:
            self.n_input_features_ = self.n_output_features_ = x[0].shape[1]
            self.measurement_matrix_ = np.eye(x[0].shape[1]).T

        self.n_consumed_samples = 0

        return self

    def transform(self, x):
        """
        Apply Identity transformation to the provided data.

        Args:
            x (array-like): The measurement data to be transformed. It must have a
                shape of (n_samples, n_input_features).

        Returns:
            array-like: Returns the transformed data which is the same as the input
                data in this case.
        """
        check_is_fitted(self, "n_input_features_")
        return x

    def get_feature_names(self, input_features=None):
        """
        Get the names of the output features.

        Args:
            input_features (list of string, optional): The string names for input
                features, if available. By default, the names "x0", "x1", ... ,
                "xn_input_features" are used.

        Returns:
            list of string: Returns the output feature names.
        """
        check_is_fitted(self, "n_input_features_")
        if input_features is None:
            input_features = [f"x{i}" for i in range(self.n_input_features_)]
        else:
            if len(input_features) != self.n_input_features_:
                raise ValueError(
                    "input_features must have n_input_features_ "
                    f"({self.n_input_features_}) elements"
                )
        return input_features



"""Module for pruning Koopman models."""
from __future__ import annotations

import numpy as np
from pykoopman.koopman import Koopman
from sklearn.utils.validation import check_is_fitted


class PrunedKoopman:
    """Prune the given original Koopman `model` at `sweep_index`.

    Parameters:
        model (Koopman): An instance of `pykoopman.koopman.Koopman`.
        sweep_index (np.ndarray): Selected indices in the original Koopman model.
        dt (float): Time step used in the original model.

    Attributes:
        sweep_index (np.ndarray): Selected indices in the original Koopman model.
        lamda_ (np.ndarray): Diagonal matrix that contains the selected eigenvalues.
        original_model (Koopman): An instance of `pykoopman.koopman.Koopman`.
        W_ (np.ndarray): Matrix that maps selected Koopman eigenfunctions back to the
            system state.

    Methods:
        fit(x): Fit the pruned model to the training data `x`.
        predict(x): Predict the system state at the next time stamp given `x`.
        psi(x_col): Evaluate the selected eigenfunctions at a given state `x`.
        phi(x_col): **Not implemented**.
        ur: **Not implemented**.
        A: **Not implemented**.
        B: **Not implemented**.
        C: Property. Returns `NotImplementedError`.
        W: Property. Returns the matrix that maps the selected Koopman eigenfunctions
            back to the system state.
        lamda: Property. Returns the diagonal matrix of selected eigenvalues.
        lamda_array: Property. Returns the selected eigenvalues as a 1D array.
        continuous_lamda_array: Property. Returns the selected eigenvalues in
            continuous-time as a 1D array.
    """

    def __init__(self, model: Koopman, sweep_index: np.ndarray, dt):
        # construct lambda
        self.sweep_index = sweep_index
        # self.lamda_ = np.diag(np.diag(model.lamda)[self.sweep_index])
        self.original_model = model
        self.time = {"dt": dt}

        # no support for controllable for now
        if self.original_model.n_control_features_ > 0:
            raise NotImplementedError

        self.A_ = None

    def fit(self, x):
        """Fit the pruned model given data matrix `x`

        Parameters
        ----------
        x : numpy.ndarray
            Training data for refitting the Koopman V

        Returns
        -------
        self : PrunedKoopman
        """

        # pruned V
        selected_eigenphi = self.psi(x.T).T
        result = np.linalg.lstsq(selected_eigenphi, x)
        # print('refit residual = {}'.format(result[1]))
        self.W_ = result[0].T

        # lamda, W = np.linalg.eig(self.original_model.A)

        self.lamda_ = np.diag(np.diag(self.original_model.lamda)[self.sweep_index]) + 0j
        # evecs = self.original_model._regressor_eigenvectors

        return self

    def predict(self, x):
        """Predict system state at the next time stamp given `x`

        Parameters
        ----------
        x : numpy.ndarray
            System state `x` in row-wise

        Returns
        -------
        xnext : numpy.ndarray
            System state at the next time stamp
        """

        if x.ndim == 1:
            x = x.reshape(1, -1)
        gnext = self.lamda @ self.psi(x.T)
        # xnext = self.compute_state_from_psi(gnext)
        xnext = self.W @ gnext
        return np.real(xnext.T)

    def psi(self, x_col):
        """Evaluate the selected psi at given state `x`

        Parameters
        ----------
        x : numpy.ndarray
            System state `x` in column-wise

        Returns
        -------
        eigenphi : numpy.ndarray
            Selected eigenfunctions' value at given state `x`
        """

        # eigenphi_ori = self.original_model.psi(x_col).T
        # eigenphi_selected = eigenphi_ori[:, self.sweep_index]

        eigenphi_ori = self.original_model.psi(x_col)
        eigenphi_selected = eigenphi_ori[self.sweep_index]
        return eigenphi_selected

    def phi(self, x_col):
        # return self.original_model._regressor_eigenvectors @ self.psi(x_col)
        raise NotImplementedError("Pruned model does not have `phi` but only `psi`")

    @property
    def ur(self):
        raise NotImplementedError("Pruned model does not have `ur`")

    @property
    def A(self):
        raise NotImplementedError(
            "Pruning only happen in eigen-space. So no self.A " "but only self.lamda"
        )

    @property
    def B(self):
        raise NotImplementedError(
            "Pruning only for autonomous system rather than " "controlled system"
        )

    @property
    def C(self):
        return NotImplementedError("Pruning model does not have `C`")

    @property
    def W(self):
        check_is_fitted(self, "W_")
        return self.W_

    @property
    def lamda(self):
        return self.lamda_

    @property
    def lamda_array(self):
        return np.diag(self.lamda) + 0j

    @property
    def continuous_lamda_array(self):
        check_is_fitted(self, "_pipeline")
        return np.log(self.lamda_array) / self.time["dt"]



from __future__ import annotations

from ._base import BaseRegressor
from ._base_ensemble import EnsembleBaseRegressor
from ._dmd import PyDMDRegressor
from ._dmdc import DMDc
from ._edmd import EDMD
from ._edmdc import EDMDc
from ._havok import HAVOK
from ._kdmd import KDMD
from ._nndmd import NNDMD

__all__ = [
    "PyDMDRegressor",
    "EDMD",
    "KDMD",
    "DMDc",
    "EDMDc",
    "EnsembleBaseRegressor",
    "HAVOK",
    "NNDMD",
]



"""module for implement modes analyzer for Koopman approximation"""
from __future__ import annotations

import abc

import numpy as np


class BaseAnalyzer(object):
    """Base class for Koopman model analyzer.

    Attributes:
        model (Koopman): An instance of `pykoopman.koopman.Koopman`.
        eigenfunction (Koopman.compute_eigenfunction): A function that evaluates Koopman
            psi.
        eigenvalues_cont (numpy.ndarray): Koopman lamda in continuous-time.
        eigenvalues_discrete (numpy.ndarray): Koopman lamda in discrete-time.
    """

    def __init__(self, model):
        """Initialize the BaseAnalyzer object.

        Args:
            model (Koopman): An instance of `pykoopman.koopman.Koopman`.
        """
        self.model = model
        self.eigenfunction = self.model.psi
        self.eigenvalues_cont = self.model.continuous_lamda_array
        self.eigenvalues_discrete = self.model.lamda_array

    def _compute_phi_minus_phi_evolved(self, t, validate_data_one_traj):
        """Compute the difference between psi evolved and psi observed.

        Args:
            t (numpy.ndarray): Time stamp of this validation trajectory.
            validate_data_one_traj (numpy.ndarray): Data matrix of this validation
                trajectory.

        Returns:
            list: Linear residual for each mode.
        """

        # shape of phi = (num_samples, num_modes)
        psi = self.eigenfunction(validate_data_one_traj.T).T

        linear_residual_list = []
        for i in range(len(self.eigenvalues_cont)):
            linear_residual_list.append(
                psi[:, i] - np.exp(self.eigenvalues_cont[i] * t) * psi[0:1, i]
            )
        return linear_residual_list

    def validate(self, t, validate_data_one_traj):
        """Validate Koopman psi.

        Given a single trajectory, compute the norm of the difference
        between observed psi and evolved psi for each mode.

        Args:
            t (numpy.ndarray): Time stamp of this validation trajectory.
            validate_data_one_traj (numpy.ndarray): Data matrix of this validation
                trajectory.

        Returns:
            list: Difference in norm for each mode.
        """

        linear_residual_list = self._compute_phi_minus_phi_evolved(
            t, validate_data_one_traj
        )
        linear_residual_norm_list = [
            np.linalg.norm(tmp) for tmp in linear_residual_list
        ]
        return linear_residual_norm_list

    @abc.abstractmethod
    def prune_model(self, *params, **kwargs):
        """Prune the model.

        This method should be implemented by the derived classes.

        Args:
            *params: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Raises:
            NotImplementedError: If the method is not implemented by the derived class.
        """
        raise NotImplementedError
